/*
 * usbgem.c: General USB to Fast Ethernet mac driver framework
 *
 * Copyright (c) 2002-2007 Masayuki Murayama.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer. 
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * 3. Neither the name of the author nor the names of its contributors may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#pragma ident	"@(#)usbgem.c	1.3 07/04/22"

/*
 Change log
 */

/*
 TODO:
	implement DELAYED_START
 */

/*
 * System Header files.
 */
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/debug.h>
#include <sys/stropts.h>
#include <sys/stream.h>
#include <sys/strlog.h>
#include <sys/kmem.h>
#include <sys/stat.h>
#include <sys/kstat.h>
#include <sys/vtrace.h>
#include <sys/dlpi.h>
#include <sys/strsun.h>
#include <sys/ethernet.h>
#include <sys/modctl.h>
#include <sys/errno.h>
#include <sys/dditypes.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/gld.h>

#include <sys/usb/usba.h>

/* supplement definitions */
extern usb_ep_data_t *usb_get_ep_data(
	dev_info_t *dip,
	usb_client_dev_data_t *dev_datap,
        uint_t interface,
        uint_t alternate,
        uint_t type,
        uint_t direction);

extern char *usb_str_cr(int);

#include "mii.h"
#include "usbgem.h"

#ifdef MODULE
char	ident[] = "usb general ethernet mac driver v" VERSION;
char	_depends_on[] = {"misc/gld"};
#endif

/* Debugging support */
#ifdef USBGEM_DEBUG_LEVEL
static int usbgem_debug = USBGEM_DEBUG_LEVEL;
#define	DPRINTF(n, args)	if (usbgem_debug>(n)) cmn_err args
#else
#define	DPRINTF(n, args)
#endif

/*
 * Useful macros and typedefs
 */
#ifndef B_FALSE
#  define	B_FALSE			(0)
#endif
#ifndef B_TRUE
#  define	B_TRUE			(!B_FALSE)
#endif
#define	ROUNDUP(x, a)		(((x) + (a) - 1) & ~((a) - 1))
#define	DEFAULT_PIPE(dp)	((dp)->reg_data->dev_default_ph)

/*
 * configuration parameters
 */
#define	USBDRV_MAJOR_VER	2
#define	USBDRV_MINOR_VER	0

#define	MAXPKTLEN(dp)	((dp)->mtu + 14)
#define	MAXPKTBUF(dp)	((dp)->mtu + 14 + ETHERFCSL)

#define	WATCH_INTERVAL_FAST	drv_usectohz(100*1000)
#undef	BUG_FREE_BULK_REQ
/*
 * Macros to distinct chip generation.
 */

/*
 * Private functions
 */
static void usbgem_intr_normal_cb(usb_pipe_handle_t, usb_intr_req_t *);
static void usbgem_intr_exc_cb(usb_pipe_handle_t, usb_intr_req_t *);

static void usbgem_bulkin_normal_cb(usb_pipe_handle_t, usb_bulk_req_t *);
static void usbgem_bulkin_exc_cb(usb_pipe_handle_t, usb_bulk_req_t *);

static void usbgem_bulkout_normal_cb(usb_pipe_handle_t, usb_bulk_req_t *);
static void usbgem_bulkout_exc_cb(usb_pipe_handle_t, usb_bulk_req_t *);

static int usbgem_mii_start(struct usbgem_dev *);
static void usbgem_mii_stop(struct usbgem_dev *);

/* local buffer management */
static void usbgem_init_tx_buf(struct usbgem_dev *);
static int usbgem_init_rx_buf(struct usbgem_dev *);
static void usbgem_tx_timeout(struct usbgem_dev *);

/* GLD interfaces */
static int usbgem_reset(gld_mac_info_t *);
static int usbgem_start(gld_mac_info_t *);
static int usbgem_stop(gld_mac_info_t *);
static int usbgem_set_mac_address(gld_mac_info_t *, uint8_t *);
static int usbgem_set_multicast(gld_mac_info_t *, uint8_t *, int);
static int usbgem_set_promiscuous(gld_mac_info_t *, int);
static int usbgem_get_stats(gld_mac_info_t *, struct gld_stats *);
static int usbgem_send(gld_mac_info_t *, mblk_t *);

static	struct ether_addr	etherbroadcastaddr = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};
static int usbgem_speed_value[] = {10, 100, 1000};

static int usbgem_ctrl_retry = 5;

/* event support */
static int usbgem_disconnect_cb(dev_info_t *dip);
static int usbgem_reconnect_cb(dev_info_t *dip);
static int usbgem_pre_suspend_event_cb(dev_info_t *dip);
static int usbgem_post_resume_event_cb(dev_info_t *dip);
#ifdef notdef
static usb_event_t usbgem_events = {
	usbgem_disconnect_cb,
	usbgem_reconnect_cb,
	usbgem_pre_suspend_event_cb,
	usbgem_post_resume_event_cb,
};
#endif

#ifdef MODULE
extern struct mod_ops mod_miscops;

static struct modlmisc modlmisc = {
	&mod_miscops,
	"usbgem v" VERSION,
};

static struct modlinkage modlinkage = {
	MODREV_1, &modlmisc, NULL
};

/*
 * _init : done
 */
int
_init(void)
{
	int 	status;

	DPRINTF(2, (CE_CONT, "!usbgem: _init: called"));
	status = mod_install(&modlinkage);

	return (status);
}

/*
 * _fini : done
 */
int
_fini(void)
{
	int	status;

	DPRINTF(2, (CE_CONT, "!usbgem: _fini: called"));
	status = mod_remove(&modlinkage);
	return (status);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}
#endif /* MODULE */

/* ============================================================== */
/*
 * gcc3 runtime routines
 */
/* ============================================================== */
#pragma weak memcmp
int
memcmp(const void *s1, const void *s2, size_t n)
{
	int	i;
	int	ret;

	ret = 0;
	for (i = 0; i < n; i++) {
		ret = (int)((uint8_t *)s1)[i] - (int)((uint8_t *)s2)[i];
		if (ret != 0) {
			return (ret);
		}
	}
	return (0);
}

#pragma weak memset
void *
memset(void *s, int c, size_t n)
{
	if ((c & 0xff) == 0) {
		bzero(s, n);
	}
	else {
		while (n--) {
			((uint8_t *)s)[n] = c;
		}
	}
	return (s);
}

#pragma weak memcpy
void *
memcpy(void *s1, const void *s2, size_t n)
{
	bcopy(s2, s1, n);
	return (s1);
}

/* ============================================================== */
/*
 * Ether CRC calculation utilities
 */
/* ============================================================== */
/*
 * Ether CRC calculation according to 21143 data sheet
 */
#define	CRC32_POLY_LE	0xedb88320
uint32_t
usbgem_ether_crc_le(uint8_t *addr)
{
	int		idx;
	int		bit;
	u_int		data;
	uint32_t	crc = 0xffffffff;

	crc = 0xffffffff;
	for (idx = 0; idx < ETHERADDRL; idx++) {
		for (data = *addr++, bit = 0; bit < 8; bit++, data >>= 1) {
			crc = (crc >> 1)
			    ^ (((crc ^ data) & 1) ? CRC32_POLY_LE : 0);
		}
	}
	return	crc;
}

#define	CRC32_POLY_BE	0x04c11db7
uint32_t
usbgem_ether_crc_be(uint8_t *addr)
{
	int		idx;
	int		bit;
	u_int		data;
	uint32_t	crc;

	crc = 0xffffffff;
	for (idx = 0; idx < ETHERADDRL; idx++) {
                for (data = *addr++, bit = 0; bit < 8; bit++, data >>= 1) {
			crc = (crc << 1)
			    ^ ((((crc >> 31) ^ data) & 1) ? CRC32_POLY_BE : 0);
		}
	}
	return crc;
}
/* ============================================================== */
/*
 * USB pipe management
 */
/* ============================================================== */
static boolean_t
usbgem_rx_start_unit(struct usbgem_dev *dp, usb_bulk_req_t *req)
{
	mblk_t		*mp;
	usb_flags_t	flags;
	int		err;

	if (req == NULL) {
		req = usb_alloc_bulk_req(dp->dip, 0, USB_FLAGS_SLEEP);
	}
	if (req == NULL) {
		cmn_err(CE_WARN,
			"!%s: usbgem_rx_start_unit: failed to allocate bulkreq",
			dp->name);
		goto err;
	}

	mp = allocb(dp->rx_buf_len, BPRI_MED);
	if (mp == NULL) {
		cmn_err(CE_WARN,
			"!%s: usbgem_rx_start_unit: failed to allocate mblk",
			dp->name);
		goto err;
	}

	req->bulk_len		= dp->rx_buf_len;
	req->bulk_data		= mp;
	req->bulk_client_private = (usb_opaque_t) dp;
	req->bulk_timeout	= 0;
	req->bulk_attributes	= USB_ATTRS_SHORT_XFER_OK;
	req->bulk_cb		= usbgem_bulkin_normal_cb;
	req->bulk_exc_cb	= usbgem_bulkin_exc_cb;
	req->bulk_completion_reason = 0;
	req->bulk_cb_flags	= 0; 

	flags = 0;
	err = usb_pipe_bulk_xfer(dp->bulkin_pipe, req, flags);

	if (err != USB_SUCCESS) {
		cmn_err(CE_WARN, "%s: failed to bulk_xfer for rx, err:%d",
			dp->name, err);
		usb_free_bulk_req(req);
		goto err;
	}

	dp->rx_line_busy++;

	return B_TRUE;
err:
	return B_FALSE;
}

/* ============================================================== */
/*
 * Rx/Tx buffer management
 */
/* ============================================================== */
static void
usbgem_init_tx_buf(struct usbgem_dev *dp)
{
	DPRINTF(0, (CE_CONT, "!%s: usbgem_init_tx_buf", dp->name));

	dp->tx_head = 0;
	dp->tx_tail = 0;
#ifdef TX_FIFO_CONTROL
	dp->tx_curr = 0;
#endif
}

static int
usbgem_init_rx_buf(struct usbgem_dev *dp)
{
	int	i;

	for (i = 0; i < dp->ugc.usbgc_rx_list_max; i++) {
		if (!usbgem_rx_start_unit(dp, NULL)) {
			return USB_FAILURE;
		}
	}

	return USB_SUCCESS;
}

#ifdef notdef
void
usbgem_restart_tx(struct usbgem_dev *dp)
{
	ASSERT(mutex_owned(&dp->txlock));

	DPRINTF(2, (CE_CONT, "!%s: usbgem_restart_tx: called: tx_line_busy:%d",
		dp->name, dp->tx_tail - dp->tx_head));

	usb_pipe_reset(dp->dip, dp->bulkout_pipe, USB_FLAGS_SLEEP, NULL, 0);

	dp->nic_state = NIC_INACTIVE;
	while (dp->tx_tail - dp->tx_head > 0) {
		cv_wait(&dp->drain_cv, &dp->txlock);
	}

	/* Reset the chip. */
	if (USBGEM_RESET_CHIP(dp) != USB_SUCCESS) {
		goto usberr;
	}
	/* Reinitialize the hardware. */
	if (USBGEM_INIT_CHIP(dp) != USB_SUCCESS) {
		goto usberr;
	}
	/* Set mac address and rx mode */
	if (USBGEM_SET_RX_FILTER(dp) != USB_SUCCESS) {
		goto usberr;
	}
	/* Restart the nic */
	dp->nic_state = NIC_ACTIVE;
	if (USBGEM_START_CHIP(dp) != USB_SUCCESS) {
		goto usberr;
	}
	return;
usberr:
	return;
}
#endif
/* ============================================================== */
/*
 * memory resource management
 */
/* ============================================================== */
static int
usbgem_alloc_memory(struct usbgem_dev* dp)
{
	int	i;

	/* allocate tx requst structure */
#ifndef BUG_FREE_BULK_REQ
	for (i = 0; i < dp->ugc.usbgc_tx_list_max; i++) {
		if ((dp->tx_list[i].req =
		     usb_alloc_bulk_req(dp->dip, 0, USB_FLAGS_SLEEP)) == NULL) {
			while (i--) {
				usb_free_bulk_req(dp->tx_list[i].req);
			}
			return USB_FAILURE;
		}
	}
#endif
	return USB_SUCCESS;
}

static int
usbgem_free_memory(struct usbgem_dev* dp)
{
	/* free tx requst structure */
	int	i;
#ifndef BUG_FREE_BULK_REQ
	/* allocate tx requst structure */
	for (i = 0; i < dp->ugc.usbgc_tx_list_max; i++) {
		usb_free_bulk_req(dp->tx_list[i].req);
	}
#endif
	return USB_SUCCESS;
}

/* ============================================================== */
/*
 * GLD interface
 */
/* ============================================================== */

static int
usbgem_reset(gld_mac_info_t *macinfo)
{
	struct usbgem_dev *dp;

	dp = (struct usbgem_dev *)macinfo->gldm_private;

	DPRINTF(3, (CE_CONT, "!%s: usbgem_reset: called", dp->name));
#ifdef notdef
	/* workaround */
	usb_pipe_reset(dp->dip, dp->bulkin_pipe, USB_FLAGS_SLEEP, NULL, 0);
#endif
	mutex_enter(&dp->rxlock);
	mutex_enter(&dp->txlock);

	usbgem_init_tx_buf(dp);

	dp->nic_state   = NIC_INACTIVE;
	dp->tx_blocked  = B_FALSE;

	dp->rx_line_busy= 0;
	mutex_exit(&dp->txlock);
	mutex_exit(&dp->rxlock);

	if (USBGEM_INIT_CHIP(dp) != USB_SUCCESS) {
		goto usberr;
	}

	return GLD_SUCCESS;

usberr:
	return GLD_FAILURE;
}

static int
usbgem_start(gld_mac_info_t *macinfo)
{
	struct usbgem_dev	*dp;
	int			err;
	usb_flags_t		flags;
	int			i;
	usb_intr_req_t		*req;
#ifdef USBGEM_DEBUG_LEVEL
	usb_pipe_state_t	p_state;
#endif
	dp = (struct usbgem_dev *)macinfo->gldm_private;

	DPRINTF(1, (CE_CONT, "!%s: usbgem_start: called", dp->name));

	if (USBGEM_START_CHIP(dp) != USB_SUCCESS) {
		cmn_err(CE_NOTE,
			"%s: usbgem_start: usb error detected", dp->name);
		goto usberr;
	}

	mutex_enter(&dp->rxlock);
	mutex_enter(&dp->txlock);
	/* init driver state */
	dp->nic_state = NIC_ACTIVE;
	mutex_exit(&dp->txlock);
	mutex_exit(&dp->rxlock);

#ifdef USBGEM_DEBUG_LEVEL
	usb_pipe_get_state(dp->intr_pipe, &p_state, 0);
	ASSERT(p_state == USB_PIPE_STATE_IDLE);
#endif /* USBGEM_DEBUG_LEVEL */

	if ((dp->misc_flag & USBGEM_NOINTR) == 0) {
		/* XXX make request */

		req = usb_alloc_intr_req(dp->dip, 0, USB_FLAGS_SLEEP);
		if (req == NULL) {
			cmn_err(CE_WARN,
				"%s: usbgem_start: failed to allocate intreq",
				dp->name);
			goto usberr;
		}
		req->intr_data		= NULL;
		req->intr_client_private = (usb_opaque_t) dp;
		req->intr_timeout	= 0;
		req->intr_attributes	= 0;
		req->intr_len		= dp->ep_intr->wMaxPacketSize;
		req->intr_cb		= usbgem_intr_normal_cb;
		req->intr_exc_cb	= usbgem_intr_exc_cb;
		req->intr_completion_reason = 0;
		req->intr_cb_flags	= 0; 

		flags = 0;
		err = usb_pipe_intr_xfer(dp->intr_pipe, req, flags);
		if (err != USB_SUCCESS) {
			cmn_err(CE_WARN,
			"%s: err:%d failed to start polling of intr pipe",
				dp->name, err);
			goto usberr;
		}
	}

	/* kick to receive first packet */
	if (usbgem_init_rx_buf(dp) != USB_SUCCESS) {
		goto usberr;
	}

	dp->timeout_id = timeout((void (*)(void *))usbgem_tx_timeout,
				(void *)dp, dp->ugc.usbgc_tx_timeout_interval);

	return GLD_SUCCESS;
usberr:
	/* stop interrupt pipe */
	if ((dp->misc_flag & USBGEM_NOINTR) == 0) {
		usb_pipe_stop_intr_polling(dp->intr_pipe, USB_FLAGS_SLEEP);
	}
	return GLD_FAILURE;
}

static int
usbgem_stop(gld_mac_info_t *macinfo)
{
	timeout_id_t		old_id;
	struct usbgem_dev	*dp;
	int			i;
	boolean_t		tx_sched;

	dp = (struct usbgem_dev *)macinfo->gldm_private;

	/* stop tx timeout watcher */
	if (dp->timeout_id) {
		do {
			untimeout(old_id = dp->timeout_id);
		} while (dp->timeout_id != old_id);	
		dp->timeout_id = 0;
	}

	/* stop preparing new rx packets and sending new packets */
	mutex_enter(&dp->rxlock);
	mutex_enter(&dp->txlock);
	if (dp->nic_state != NIC_DISCONNECTED) {
		dp->nic_state = NIC_INACTIVE;
	}
	mutex_exit(&dp->txlock);
	mutex_exit(&dp->rxlock);

	/* stop polling interrupt pipe */
	if ((dp->misc_flag & USBGEM_NOINTR) == 0) {
		usb_pipe_stop_intr_polling(dp->intr_pipe, USB_FLAGS_SLEEP);
	}

	/* cancel pended rx and tx packets */
	if (dp->nic_state != NIC_DISCONNECTED) {
		/*
		 * XXX - calling usb_pipe_reset causes to panic the system
		 * while the usb device is disconnected.
		 */
		usb_pipe_reset(dp->dip, dp->bulkin_pipe,
				USB_FLAGS_SLEEP, NULL, 0);
		usb_pipe_reset(dp->dip, dp->bulkout_pipe,
				USB_FLAGS_SLEEP, NULL, 0);
		/* stop nic hardware */
		(void) USBGEM_STOP_CHIP(dp);
	}

	DPRINTF(1, (CE_CONT,
		"!%s: usbgem_stop: rx_line_busy:%d tx_line_busy:%d",
		dp->name, dp->rx_line_busy, dp->tx_tail - dp->tx_head));

	/* ensure calback routines have been executed for canceled packets */
	mutex_enter(&dp->rxlock);
	while (dp->rx_line_busy > 0) {
		cv_wait(&dp->drain_cv, &dp->rxlock);
	}
	mutex_exit(&dp->rxlock);
	mutex_enter(&dp->txlock);
	while (dp->tx_tail - dp->tx_head > 0) {
		cv_wait(&dp->drain_cv, &dp->txlock);
	}
	mutex_exit(&dp->txlock);

	return GLD_SUCCESS;
}

static int
usbgem_set_multicast(gld_mac_info_t *macinfo, uint8_t *ep, int flag)
{
	struct usbgem_dev	*dp;
	size_t			len;
	int			i;
	int			cnt;

	dp = (struct usbgem_dev *)macinfo->gldm_private;

	cnt = dp->mc_count;

	if (flag == GLD_MULTI_ENABLE) {
		/* append new addess into mclist */
		if (dp->mc_count_req++ < USBGEM_MAXMC) {
			cnt = dp->mc_count;
			bcopy(ep, dp->mc_list[cnt].addr.ether_addr_octet,
				ETHERADDRL);
			if (dp->ugc.usbgc_multicast_hash != NULL) {
				dp->mc_list[cnt].hash =
					USBGEM_MULTICAST_HASH(dp, ep);
			}
			dp->mc_count += cnt + 1;
		}
	}
	else {
		dp->mc_count_req--;
		cnt = dp->mc_count;
		for (i = 0; i < cnt; i++) {
			if (bcmp(ep, &dp->mc_list[i].addr, ETHERADDRL) != 0) {
				continue;
			}
			/* shrink mclist table by copying forward */
			len = (cnt - (i + 1)) * sizeof(*dp->mc_list);
			if (len > 0) {
				bcopy(&dp->mc_list[i+1], &dp->mc_list[i], len);
			}
			dp->mc_count--;
		}
	}
x:
	dp->rxmode &= ~RXMODE_MULTI_OVF;
	if (dp->mc_count_req != dp->mc_count) {
		/* multicast address list overflow */
		dp->rxmode |= RXMODE_MULTI_OVF;
	}

	if (USBGEM_SET_RX_FILTER(dp) != USB_SUCCESS) {
		goto usberr;
	}

	return GLD_SUCCESS;

usberr:
	return GLD_FAILURE;
}

static int
usbgem_set_promiscuous(gld_mac_info_t *macinfo, int flag)
{
	struct usbgem_dev *dp;

	dp = (struct usbgem_dev *)macinfo->gldm_private;

	/* start serialization */
	if (flag == GLD_MAC_PROMISC_NONE) {
		dp->rxmode &= ~(RXMODE_PROMISC | RXMODE_ALLMULTI);
	}
	else if (flag == GLD_MAC_PROMISC_MULTI) {
		dp->rxmode |= RXMODE_ALLMULTI;
	}
	else {
		dp->rxmode |= RXMODE_PROMISC;
	}
	if (USBGEM_SET_RX_FILTER(dp) != USB_SUCCESS) {
		goto usberr;
	}
	/* end serialization */

	return GLD_SUCCESS;

usberr:
	return GLD_FAILURE;
}

static int
usbgem_set_mac_address(gld_mac_info_t *macinfo, uint8_t *mac) /* done */
{
	struct usbgem_dev	*dp;

	dp = (struct usbgem_dev *)macinfo->gldm_private;

	/* start serialization */
	bcopy(mac, dp->cur_addr.ether_addr_octet, ETHERADDRL);

	if (USBGEM_SET_RX_FILTER(dp) != USB_SUCCESS) {
		goto usberr;
	}
	/* end serialization */

	return GLD_SUCCESS;

usberr:
	return GLD_FAILURE;
}

static	int
usbgem_get_stats(gld_mac_info_t *macinfo, struct gld_stats *gs)
{
	struct usbgem_dev		*dp;
	struct usbgem_stats	*vs;

	dp = (struct usbgem_dev *)macinfo->gldm_private;

	/* start serialization */
	if (USBGEM_GET_STATS_CHIP(dp) != USB_SUCCESS) {
		goto usberr;
	}

	vs = &dp->stats;

	gs->glds_errxmt    = vs->errxmt;
	gs->glds_errrcv    = vs->errrcv;
	gs->glds_collisions= vs->collisions;

	gs->glds_excoll    = vs->excoll;
	gs->glds_defer     = vs->defer;
	gs->glds_frame     = vs->frame;
	gs->glds_crc       = vs->crc;

	gs->glds_overflow  = vs->overflow; /* fifo err,underrun,rbufovf*/
	gs->glds_underflow = vs->underflow;
	gs->glds_short     = vs->runt;
	gs->glds_missed    = vs->missed; /* missed pkts while rbuf ovf */
	gs->glds_xmtlatecoll = vs->xmtlatecoll;
	gs->glds_nocarrier = vs->nocarrier;
	gs->glds_norcvbuf  = vs->norcvbuf;	/* OS resource exaust */
	gs->glds_intr      = vs->intr;

	/* all before here must be kept in place for v0 compatibility */
	gs->glds_speed     = usbgem_speed_value[dp->speed] * 1000000;
	gs->glds_media     = GLDM_PHYMII;
	gs->glds_duplex    = dp->full_duplex
				? GLD_DUPLEX_FULL : GLD_DUPLEX_HALF;

	/* gs->glds_media_specific */
	gs->glds_dot3_first_coll     = vs->first_coll;
	gs->glds_dot3_multi_coll     = vs->multi_coll;
	gs->glds_dot3_sqe_error	     = 0;
	gs->glds_dot3_mac_xmt_error  = 0;
	gs->glds_dot3_mac_rcv_error  = 0;
	gs->glds_dot3_frame_too_long = vs->frame_too_long;
	/* end serialization */

	return GLD_SUCCESS;

usberr:
	return GLD_FAILURE;
}

/*==========================================================*/
/*
 * Start transmission.
 * Return zero on success,
 */
/*==========================================================*/

#ifdef TXTIMEOUT_TEST
static int usbgem_send_cnt = 0;
#endif

/*
 * usbgem_send is used only to send data packet into ethernet line.
 */

static int
usbgem_send(gld_mac_info_t *macinfo, mblk_t *mp)
{
	struct usbgem_dev	*dp;
	int			err;
	int			timeout;
	usb_flags_t		flags;
	mblk_t			*new;
	struct txbuf		*txbp;
	usb_bulk_req_t		*req;
#ifdef USBGEM_DEBUG_LEVEL
	usb_pipe_state_t	p_state;
#endif
	dp = (struct usbgem_dev *)macinfo->gldm_private;

	DPRINTF(2, (CE_CONT, "!%s: usbgem_send: called", dp->name));

	new = USBGEM_TX_MAKE_PKT(dp, mp);

	mutex_enter(&dp->txlock);

	/*
	 * Avoid to send packets while the link or nic is down.
	 */
	if (dp->mii_state != MII_STATE_LINKUP || dp->nic_state != NIC_ACTIVE) {
		/*
		 * the device is not working
		 */
		dp->stats.errxmt++;
		mutex_exit(&dp->txlock);

		if (new != NULL && new != mp) {
			/* the messages was reallocated, free it. */
			freemsg(new);
		}

		DPRINTF(2, (CE_CONT,
		"!%s: usbgem_send: failure: nic_state:%d, link_state:%d",
				dp->name, dp->nic_state, dp->mii_state));
		return GLD_NOLINK;
	}

	if (new == NULL ||
	    dp->tx_tail - dp->tx_head >= dp->ugc.usbgc_tx_list_max) {
		/*
		 * tx list is full or no memory resource.
		 */
		dp->tx_blocked = B_TRUE;
		mutex_exit(&dp->txlock);

		if (new != NULL && new != mp) {
			/* free reallocated message. */
			freemsg(new);
		}

		DPRINTF(0, (CE_CONT, "!%s: usbgem_send: no free slot",
				dp->name));
		return GLD_NORESOURCES;
	}

	DPRINTF(2, (CE_CONT, "!%s: usbgem_send: sending", dp->name));

#ifdef notdef /*USBGEM_DEBUG_LEVEL*/
	usb_pipe_get_state(dp->bulkout_pipe, &p_state, 0);
	ASSERT(p_state == USB_PIPE_STATE_ACTIVE);
#endif
	ASSERT(new->b_cont == NULL);
	dp->tx_blocked = B_FALSE;
	txbp = &dp->tx_list[SLOT(dp, dp->tx_tail)];
#ifdef BUG_FREE_BULK_REQ
	req = usb_alloc_bulk_req(dp->dip, 0, USB_FLAGS_SLEEP);
#else
	req  = txbp->req;
#endif
	req->bulk_len		= new->b_wptr - new->b_rptr;
	req->bulk_data		= new;
	req->bulk_client_private = (usb_opaque_t) dp;
	req->bulk_timeout	= 10;	/* in second */
	req->bulk_attributes	= 0;
	req->bulk_cb		= usbgem_bulkout_normal_cb;
	req->bulk_exc_cb	= usbgem_bulkout_exc_cb;
	req->bulk_completion_reason = 0;
	req->bulk_cb_flags	= 0; 

	if ((err = usb_pipe_bulk_xfer(dp->bulkout_pipe, req, 0))
							!= USB_SUCCESS) {
		mutex_exit(&dp->txlock);

		cmn_err(CE_WARN,
		"%s: usbgem_send: usb_pipe_bulk_xfer: failed: err:%d",
			dp->name, err);

		err =  GLD_FAILURE;
		goto x;
        }

	/* append this mp at the tail of the tx list */
	txbp->mp = new;
	txbp->stime = ddi_get_lbolt();

	dp->tx_tail++;
	mutex_exit(&dp->txlock);

	err = GLD_SUCCESS;
x:
	if (new != mp) {
		freemsg(mp);
	}

	if (dp->nic_state != NIC_ACTIVE) {
		/* wakeup waiting thread */
		cv_signal(&dp->drain_cv);
	}

	return err;
}

static void
usbgem_tx_timeout(struct usbgem_dev *dp)
{
	clock_t		now;
	boolean_t	tx_sched;

	now = ddi_get_lbolt();
	tx_sched = B_FALSE;

	mutex_enter(&dp->txlock); 

	switch (dp->nic_state) {
	case NIC_ERROR:
		break;

	case NIC_ACTIVE:
		if (dp->mii_state != MII_STATE_LINKUP ||
		    dp->tx_tail - dp->tx_head == 0) {
			mutex_exit(&dp->txlock); 
			goto schedule_next;
		}

		tx_sched = dp->tx_blocked;

		if (now - dp->tx_list[SLOT(dp, dp->tx_head)].stime
					< dp->ugc.usbgc_tx_timeout) {
			mutex_exit(&dp->txlock); 
			goto schedule_next;
		}
		break;

	default:
		mutex_exit(&dp->txlock); 
		goto schedule_next;
	}

	cmn_err(CE_WARN, "%s: tx timeout: nic_state:%d starttime:%d",
		dp->name, dp->nic_state, dp->tx_start_time);

	mutex_exit(&dp->txlock); 

	/* stop preparing and sending packets */
	mutex_enter(&dp->rxlock); 
	mutex_enter(&dp->txlock); 
	if (dp->nic_state != NIC_DISCONNECTED) {
		dp->nic_state = NIC_INACTIVE;
	}
	else {
		mutex_exit(&dp->txlock); 
		mutex_exit(&dp->rxlock); 
		goto schedule_next;
	}
	mutex_exit(&dp->txlock); 
	mutex_exit(&dp->rxlock); 

	/*
	 * Reset all pipes, USB_SLEEP flags garantee all request return.
	 * it implies that pended tx and rx packets will be discarded.
	 */
	usb_pipe_reset(dp->dip, dp->bulkin_pipe, USB_FLAGS_SLEEP, NULL, 0);
	usb_pipe_reset(dp->dip, dp->bulkout_pipe, USB_FLAGS_SLEEP, NULL, 0);
	usb_pipe_reset(dp->dip, dp->intr_pipe, USB_FLAGS_SLEEP, NULL, 0);

	/* ensure calback routines have been executed for canceled packets */
	mutex_enter(&dp->rxlock);
	while (dp->rx_line_busy > 0) {
		cv_wait(&dp->drain_cv, &dp->rxlock);
	}
	mutex_exit(&dp->rxlock);
	mutex_enter(&dp->txlock);
	while (dp->tx_tail - dp->tx_head > 0) {
		cv_wait(&dp->drain_cv, &dp->txlock);
	}
	mutex_exit(&dp->txlock);

	/* Reset the usb device. */
	if (USBGEM_RESET_CHIP(dp) != USB_SUCCESS) {
		goto usberr;
	}
	usbgem_init_tx_buf(dp);

	/* Reinitialize the hardware. */
	if (USBGEM_INIT_CHIP(dp) != USB_SUCCESS) {
		goto usberr;
	}

	/* Set mac address and rx mode */
	if (USBGEM_SET_RX_FILTER(dp) != USB_SUCCESS) {
		goto usberr;
	}

	/* Restart the nic */
	mutex_enter(&dp->rxlock);
	mutex_enter(&dp->txlock);

	dp->nic_state  = NIC_ACTIVE;
	dp->tx_blocked = B_FALSE;
	tx_sched       = B_TRUE;

	mutex_exit(&dp->txlock);
	mutex_exit(&dp->rxlock);

	if (USBGEM_START_CHIP(dp) != USB_SUCCESS) {
		goto usberr;
	}

	if (usbgem_init_rx_buf(dp) != USB_SUCCESS) {
		goto usberr;
	}

#ifdef TXTIMEOUT_TEST
	usbgem_send_cnt = 0;
#endif

schedule_next:
	if (tx_sched) {
		gld_sched(dp->macinfo);
	}

	DPRINTF(4, (CE_CONT, "!%s: blocked:%d", dp->name, dp->tx_blocked));
	goto next;

usberr:
	mutex_enter(&dp->rxlock);
	mutex_enter(&dp->txlock);
	if (dp->nic_state != NIC_DISCONNECTED) {
		dp->nic_state = NIC_ERROR;
		cmn_err(CE_WARN, "%s: usbgem_tx_timeout: usberr detected",
			dp->name);
	}
	mutex_exit(&dp->txlock);
	mutex_exit(&dp->rxlock);

next:
	dp->timeout_id =
		timeout((void (*)(void *))usbgem_tx_timeout,
			(void *)dp, dp->ugc.usbgc_tx_timeout_interval);
}

/*==================================================================*/
/*
 * Callback handlers
 */
/*==================================================================*/
static void
usbgem_bulkin_normal_cb(usb_pipe_handle_t pipe, usb_bulk_req_t *req)
{
	int			err;
	mblk_t			*newmp;
	uint_t			flags;
	struct usbgem_dev	*dp;
	mblk_t			*mp;

	dp = (struct usbgem_dev *)req->bulk_client_private;
	mp = req->bulk_data;
	req->bulk_data = NULL;

	DPRINTF(2, (CE_CONT, "!%s: %s: mp:%p, cr:%s(%d)",
		dp->name, __func__, mp,
		usb_str_cr(req->bulk_completion_reason),
		req->bulk_completion_reason));

	if (req->bulk_completion_reason == USB_CR_OK &&
	    dp->nic_state == NIC_ACTIVE) {
		newmp = USBGEM_RX_MAKE_PKT(dp, mp);

		if (newmp != mp) {
			/* the messages was reallocated, free old one */
			freemsg(mp);
		}

		if (newmp != NULL) {
			/* send up if it is a valid packet */
			gld_recv(dp->macinfo, newmp);
		}
	}
	else {
		freemsg(mp);
	}

	mutex_enter(&dp->rxlock);

	dp->rx_line_busy--;

	if (dp->nic_state == NIC_ACTIVE) {
		/*
		 * prepare to receive next packet
		 */
		(void) usbgem_rx_start_unit(dp, req);
	}
	else {
		usb_free_bulk_req(req);
	}

	if (dp->nic_state != NIC_ACTIVE) {
		cv_broadcast(&dp->drain_cv);
	}

	mutex_exit(&dp->rxlock);
}

static void
usbgem_bulkin_exc_cb(usb_pipe_handle_t pipe, usb_bulk_req_t *req)
{
	int			err;
	mblk_t			*newmp;
	uint_t			flags;
	struct usbgem_dev	*dp;
	mblk_t			*mp;

	dp = (struct usbgem_dev *)req->bulk_client_private;
	mp = req->bulk_data;

	DPRINTF(0, (CE_CONT,
		"!%s: %s: mp:%p cr:%s(%d) cb_flags:0x%x",
		dp->name, __func__, mp,
		usb_str_cr(req->bulk_completion_reason),
		req->bulk_completion_reason,
		req->bulk_cb_flags));

	usbgem_bulkin_normal_cb(pipe, req);

	DPRINTF(0, (CE_CONT,
		"!%s: %s: rx_line_busy:%d",
		dp->name, __func__, dp->rx_line_busy));
}

static void
usbgem_bulkout_normal_cb(usb_pipe_handle_t pipe, usb_bulk_req_t *req)
{
	int			err;
	usb_flags_t		flags;
	int			timeout;
	mblk_t			*newmp;
	struct usbgem_dev	*dp;
	mblk_t			*mp;
	boolean_t		tx_sched;

	dp = (struct usbgem_dev *)req->bulk_client_private;
	mp = req->bulk_data;
	req->bulk_data = NULL;

	DPRINTF(2, (CE_CONT,
		"!%s: %s: cr:%s(%d) cb_flags:0x%x head:%d tail:%d",
		dp->name, __func__,
		usb_str_cr(req->bulk_completion_reason),
		req->bulk_completion_reason,
		req->bulk_cb_flags,
		dp->tx_head, dp->tx_tail));

	/*
	 *  we have done to transfer a packet into tx fifo.
	 */
	mutex_enter(&dp->txlock);

	ASSERT(dp->tx_tail - dp->tx_head > 0);
	ASSERT(dp->tx_list[SLOT(dp, dp->tx_head)].mp == mp);

	dp->tx_head++;

	tx_sched = B_FALSE;
	if (dp->tx_blocked) {
		tx_sched = (dp->tx_tail - dp->tx_head <= 2);
	}

	if (dp->nic_state != NIC_ACTIVE) {
		cv_broadcast(&dp->drain_cv);
	}
	mutex_exit(&dp->txlock);
#ifdef BUG_FREE_BULK_REQ
	usb_free_bulk_req(req);
#endif
	freemsg(mp);

	if (tx_sched) {
		gld_sched(dp->macinfo);
	}
}

static void
usbgem_bulkout_exc_cb(usb_pipe_handle_t pipe, usb_bulk_req_t *req)
{
	boolean_t		send;
	int			timeout;
	int			err;
	mblk_t			*newmp;
	usb_flags_t		flags;
	struct usbgem_dev	*dp;
	mblk_t			*mp;

	dp = (struct usbgem_dev *)req->bulk_client_private;
	mp = req->bulk_data;

	DPRINTF(0, (CE_CONT,
	"!%s: usbgem_bulkout_exc_cb: cr:%s(%d) cb_flags:0x%x head:%d tail:%d",
		dp->name,
		usb_str_cr(req->bulk_completion_reason),
		req->bulk_completion_reason,
		req->bulk_cb_flags,
		dp->tx_head, dp->tx_tail));
#ifdef notdef
	if (req->bulk_completion_reason == 12) {
		/* time out */
		usb_pipe_reset(dp->dip, dp->bulkout_pipe, 0, NULL, 0);
	}
#endif
	usbgem_bulkout_normal_cb(pipe, req);
}

static void
usbgem_intr_normal_cb(usb_pipe_handle_t ph, usb_intr_req_t *req)
{
	struct usbgem_dev	*dp;

	dp = (struct usbgem_dev *)req->intr_client_private;

	if (req->intr_completion_reason == USB_CR_OK &&
	    dp->nic_state == NIC_ACTIVE) {
		USBGEM_INTERRUPT(dp, req->intr_data);
	}

	/* free the request and data */
	usb_free_intr_req(req);
}

static void
usbgem_intr_exc_cb(usb_pipe_handle_t ph, usb_intr_req_t *req)
{
	usbgem_intr_normal_cb(ph, req);
}

/* ======================================================================== */
/*
 * MII support routines
 */
/* ======================================================================== */
static uint16_t
usbgem_mii_read(struct usbgem_dev *dp, uint_t reg, int *errp)
{
	if (!dp->no_preamble) {
		USBGEM_MII_SYNC(dp, errp);
		if (*errp != USB_SUCCESS) {
			return 0;
		}
	}
	return USBGEM_MII_READ(dp, reg, errp);
}

static void
usbgem_mii_write(struct usbgem_dev *dp, uint_t reg, uint16_t val, int *errp)
{
	if (!dp->no_preamble) {
		USBGEM_MII_SYNC(dp, errp);
		if (*errp != USB_SUCCESS) {
			return;
		}
	}
	USBGEM_MII_WRITE(dp, reg, val, errp);
}

#define	fc_cap_decode(x)	\
	((((x) & MII_ABILITY_PAUSE) != 0 ? 1 : 0) |	\
	 (((x) & MII_ABILITY_ASM_DIR) != 0 ? 2 : 0))

int
usbgem_mii_config_default(struct usbgem_dev *dp, int *errp)
{
	uint16_t	mii_stat;
	uint16_t	xstat;
	uint16_t	val;
	static uint16_t fc_cap_encode[4] = {
		/* none */		0,
		/* symmetric */		MII_ABILITY_PAUSE,
		/* tx */		MII_ABILITY_ASM_DIR,
		/* rx-symmetric */	MII_ABILITY_PAUSE | MII_ABILITY_ASM_DIR,
	};

	DPRINTF(2, (CE_CONT, "!%s: gem_mii_config: called", dp->name));

	/*
	 * Set advertisement register
	 */
	mii_stat = usbgem_mii_read(dp, MII_STATUS, errp);
	DPRINTF(2, (CE_CONT, "!%s: usbgem_mii_config: MII_STATUS reg:%b",
		dp->name, mii_stat, MII_STATUS_BITS));
	if (*errp != USB_SUCCESS) {
		goto usberr;
	}

	if ((mii_stat & MII_STATUS_ABILITY) == 0) {
		/* it's funny */
		cmn_err(CE_WARN, "%s: wrong ability bits: mii_status:%b",
			dp->name, mii_stat, MII_STATUS_BITS);
		return -1;
	}

	/* Do not change the rest of ability bits in advert reg */
	val = usbgem_mii_read(dp, MII_AN_ADVERT, errp) & ~MII_ABILITY;
	if (*errp != USB_SUCCESS) {
		goto usberr;
	}

	if ((mii_stat & MII_STATUS_100_BASE_T4) != 0 && dp->anadv_100t4) {
		val |= MII_ABILITY_100BASE_T4;
	}
	if ((mii_stat & MII_STATUS_100_BASEX_FD) != 0 && dp->anadv_100fdx) {
		val |= MII_ABILITY_100BASE_TX_FD;
	}
	if ((mii_stat & MII_STATUS_100_BASEX) != 0 && dp->anadv_100hdx) {
		val |= MII_ABILITY_100BASE_TX;
	}
	if ((mii_stat & MII_STATUS_10_FD) != 0 && dp->anadv_10fdx) {
		val |= MII_ABILITY_10BASE_T_FD;
	}
	if ((mii_stat & MII_STATUS_10) != 0 && dp->anadv_10hdx) {
		val |= MII_ABILITY_10BASE_T;
	}

	/* set flow control capability */
	val |= fc_cap_encode[dp->ugc.usbgc_flow_control];

	DPRINTF(1, (CE_CONT,
		"!%s: usbgem_mii_config: "
			"setting MII_AN_ADVERT reg:%b, mii_mode:%d, fc:%d",
		dp->name, val, MII_ABILITY_BITS, dp->ugc.usbgc_mii_mode,
                dp->ugc.usbgc_flow_control));

	usbgem_mii_write(dp, MII_AN_ADVERT, val, errp);
	if (*errp != USB_SUCCESS) {
		goto usberr;
	}
#ifdef USBGEM_GIGA
	if (dp->ugc.usbgc_mii_mode == USBGEM_MODE_1000BASET) {
		/*
		 * 1000Base-T GMII support
		 */
		val = usbgem_mii_read(dp, MII_1000TC, errp) &
				~(MII_1000TC_ADV_FULL | MII_1000TC_ADV_HALF);
		if (*errp != USB_SUCCESS) {
			goto usberr;
		}
		xstat = usbgem_mii_read(dp, MII_XSTATUS, errp);
		if (*errp != USB_SUCCESS) {
			goto usberr;
		}
		if ((xstat & MII_XSTATUS_1000BASET_FD) != 0 &&
		     dp->anadv_1000fdx) {
			val |= MII_1000TC_ADV_FULL;
		}
		if ((xstat & MII_XSTATUS_1000BASET) != 0 &&
		     dp->anadv_1000hdx) {
			val |= MII_1000TC_ADV_HALF;
		}

		DPRINTF(1, (CE_CONT,
			"!%s: usbgem_mii_config: setting MII_1000TC reg:%b",
			dp->name, val, MII_1000TC_BITS));

		usbgem_mii_write(dp, MII_1000TC, val, errp);
		if (*errp != USB_SUCCESS) {
			goto usberr;
		}
	}
#endif /* USBGEM_GIGA */
	return USB_SUCCESS;

usberr:
	return *errp;
}

clock_t
usbgem_mii_link_check(struct usbgem_dev *dp)
{
	uint16_t	mii_stat;
	uint16_t	advert;
	uint16_t	lpable;
	uint16_t	exp;
	uint16_t	ctl1000;
	uint16_t	stat1000;
	uint16_t	val;
	clock_t		now;
	clock_t		diff;
	clock_t		interval;
	boolean_t	use_rsan = B_TRUE;
	int		err;
	int		ret;
	int		linkdown_action;

	ASSERT(mutex_owned(&dp->rxlock));

	now = ddi_get_lbolt();
	diff = now - dp->mii_last_check;
	dp->mii_last_check = now;

	DPRINTF(3, (CE_CONT, "!%s: mii_link_check: time:%d state:%d",
				dp->name, now, dp->mii_state));

	if (dp->nic_state == NIC_DISCONNECTED) {
		goto usberr;
	}

next_nowait:
	switch (dp->mii_state) {
	case MII_STATE_UNKNOWN:
		mii_stat = usbgem_mii_read(dp, MII_STATUS, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		goto reset_phy;

	case MII_STATE_RESETTING:
		dp->mii_timer -= diff;
		if (dp->mii_timer > 0) {
			/* wait for time-up */
			interval = WATCH_INTERVAL_FAST;
			goto next;
		}

		/* Timer expired, ensure reset bit is not set */

		USBGEM_MII_SYNC(dp, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}

		val = usbgem_mii_read(dp, MII_CONTROL, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		if ((val & (MII_CONTROL_RESET |
			    MII_CONTROL_ISOLATE | MII_CONTROL_PWRDN)) != 0) {
			cmn_err(CE_NOTE,
			"!%s: resetting phy not complete. mii_control:0x%b",
				dp->name, val);
			usbgem_mii_write(dp, MII_CONTROL, 0, &err);
			if (err != USB_SUCCESS) {
				goto usberr;
			}
		}

		/* Configure PHY registers */
		val = USBGEM_MII_CONFIG(dp, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		if (val != 0) {
			goto reset_phy;
		}

		if (dp->mii_fixedmode) {
			dp->mii_state = MII_STATE_MEDIA_SETUP;
			dp->mii_timer = 0;
			interval = 0;
			goto next_nowait;
		}

		/* Issue auto-negotiation command */
		goto autonego;

	case MII_STATE_AUTONEGOTIATING:
		/*
		 * Autonegotiation in progress
		 */
		dp->mii_timer -= diff;
		if (dp->mii_timer -
		      (dp->ugc.usbgc_mii_an_timeout - dp->ugc.usbgc_mii_an_wait) > 0) {
			/* wait for minimum time (2.3 - 2.5 sec) for rhine */
			interval = WATCH_INTERVAL_FAST;
			goto next;
		}

		/* read PHY status */
		mii_stat = usbgem_mii_read(dp, MII_STATUS, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		DPRINTF(4, (CE_CONT,
			"!%s: usbgem_mii_link_check: called: "
			"mii_state:%d MII_STATUS reg:%b",
			dp->name, dp->mii_state,
			mii_stat, MII_STATUS_BITS));

		if ((mii_stat & MII_STATUS_REMFAULT) != 0) {
			/*
			 * The link parnert told me something wrong happend.
			 * What do we do ?
			 */
			cmn_err(CE_CONT,
				"!%s: auto-negotiation failed: remote fault",
				dp->name);
			goto autonego;
		}

		if ((mii_stat & MII_STATUS_ANDONE) == 0) {
			if (dp->mii_timer <= 0) {
				/*
				 * Auto-negotiation timed out,
				 * Reset PHY and try again.
				 */
				if (!dp->mii_supress_msg) {
					cmn_err(CE_WARN,
					"!%s: auto-negotiation failed: timeout",
					dp->name);
					dp->mii_supress_msg = B_TRUE;
				}
#ifdef notdef
				/* stop auto negotiation */
				val = usbgem_mii_read(dp, MII_CONTROL, &err);
				if (err != USB_SUCCESS) {
					goto usberr;
				}
				usbgem_mii_write(dp, MII_CONTROL,
				   val & ~(MII_CONTROL_RSAN | MII_CONTROL_ANE),
				   &err);
				if (err != USB_SUCCESS) {
					goto usberr;
				}
#endif
				goto autonego;
			}
			/*
			 * Auto-negotiation is in progress. Wait.
			 */
			interval = dp->ugc.usbgc_mii_an_watch_interval;
			goto next;
		}

		/*
		 * Auto-negotiation have done.
		 * Assume linkdown and fall through.
		 */
		DPRINTF(2, (CE_CONT, "!%s: auto-negotiation done", dp->name));
		dp->mii_supress_msg = B_FALSE;
		dp->mii_state = MII_STATE_AN_DONE;
		dp->mii_timer = 0;
		diff = 0;
		goto next_nowait;

	case MII_STATE_AN_DONE:
		/*
		 * Auto-negotiation have done.
		 * Now we can set up media.
		 */
		dp->mii_timer -= diff;
		if (dp->mii_timer > 0) {
			/*
			 * wait for a while
			 */
			interval = WATCH_INTERVAL_FAST;
			goto next;
		}

		/*
		 * set up the result of auto negotiation 
		 */

		/* determine full/half and 1000Mbps/100Mbps/10Mbps */
		advert = usbgem_mii_read(dp, MII_AN_ADVERT, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		lpable = usbgem_mii_read(dp, MII_AN_LPABLE, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		exp = usbgem_mii_read(dp, MII_AN_EXPANSION, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}

		DPRINTF(0, (CE_CONT,
		"!%s: auto-negotiation done: advert:%b, lpable:%b, exp:%b",
			dp->name,
			advert, MII_ABILITY_BITS,
			lpable, MII_ABILITY_BITS,
			exp, MII_AN_EXP_BITS));
		/*
		 * configure link mode according to AN priority.
		 */	
		val = advert & lpable;
		ctl1000  = 0;
		stat1000 = 0;
		if (dp->ugc.usbgc_mii_mode == USBGEM_MODE_1000BASET) {
			ctl1000  = USBGEM_MII_READ(dp, MII_1000TC, &err);
			if (err != USB_SUCCESS) {
				goto usberr;
			}
			stat1000 = USBGEM_MII_READ(dp, MII_1000TS, &err);
			if (err != USB_SUCCESS) {
				goto usberr;
			}
			DPRINTF(0, (CE_CONT,
				"!%s: MII_1000TC reg:%b, MII_1000TS reg:%b",
				dp->name,
				ctl1000, MII_1000TC_BITS,
				stat1000, MII_1000TS_BITS));
		}

		if ((ctl1000 & MII_1000TC_ADV_FULL) != 0 &&
		    (stat1000 & MII_1000TS_LP_FULL) != 0) {
			/* 1000BaseT & full duplex */
			dp->speed	 = USBGEM_SPD_1000;
			dp->full_duplex  = B_TRUE;
		}
		else if ((ctl1000 & MII_1000TC_ADV_HALF) != 0 &&
			 (stat1000 & MII_1000TS_LP_HALF) != 0) {
			/* 1000BaseT & half duplex */
			dp->speed	 = USBGEM_SPD_1000;
			dp->full_duplex  = B_FALSE;
		}
		else if ((val & MII_ABILITY_100BASE_TX_FD) != 0) {
			/* 100BaseTx & fullduplex */
			dp->speed	 = USBGEM_SPD_100;
			dp->full_duplex  = B_TRUE;
		}
		else if ((val & MII_ABILITY_100BASE_TX) != 0) {
			/* 100BaseTx & half duplex */
			dp->speed	 = USBGEM_SPD_100;
			dp->full_duplex  = B_FALSE;
		}
		else if ((val & MII_ABILITY_10BASE_T_FD) != 0) {
			/* 10BaseT & full duplex */
			dp->speed	 = USBGEM_SPD_10;
			dp->full_duplex  = B_TRUE;
		}
		else if ((val & MII_ABILITY_10BASE_T) != 0) {
			/* 10BaseT & half duplex */
			dp->speed	 = USBGEM_SPD_10;
			dp->full_duplex  = B_FALSE;
		}
		else {
			/*
			 * It seem that the link partnar don't have
			 * auto-negotiation capability and our PHY
			 * could not report the correct current mode.
			 * This case happened on VT86C100A chip so
			 * that we guess current mode by mii_control
			 * register.
			 */

			val = usbgem_mii_read(dp, MII_CONTROL, &err);
			if (err != USB_SUCCESS) {
				goto usberr;
			}

			cmn_err(CE_NOTE,
				"!%s: auto-negotiation done but "
				"common ability not found.\n"
				"PHY state: control:%b advert:%b lpable:%b\n"
				"guessing the following mode...",
				dp->name,
				val, MII_CONTROL_BITS,
				advert, MII_ABILITY_BITS,
				lpable, MII_ABILITY_BITS);

			/* select 100m full or 10m half */
			dp->speed = ((val & MII_CONTROL_100MB) != 0)
					? USBGEM_SPD_100 : USBGEM_SPD_10;
			dp->full_duplex  = dp->speed != USBGEM_SPD_10;
		}

		if (dp->full_duplex) {
			static uint8_t fc_result[4/*my cap*/][4/*lp cap*/] = {
			/*	 none	symm	tx	rx/symm */
			/* none */
				{FLOW_CONTROL_NONE,
					FLOW_CONTROL_NONE,
						FLOW_CONTROL_NONE,
							FLOW_CONTROL_NONE},
			/* sym */
				{FLOW_CONTROL_NONE,
					FLOW_CONTROL_SYMMETRIC,
						FLOW_CONTROL_NONE,
							FLOW_CONTROL_SYMMETRIC},
			/* tx */
				{FLOW_CONTROL_NONE,
					FLOW_CONTROL_NONE,
						FLOW_CONTROL_NONE,
							FLOW_CONTROL_TX_PAUSE},
			/* rx/symm */
				{FLOW_CONTROL_NONE,
					FLOW_CONTROL_SYMMETRIC,
						FLOW_CONTROL_RX_PAUSE,
							FLOW_CONTROL_SYMMETRIC},
			};
			dp->flow_control =
			fc_result[fc_cap_decode(advert)][fc_cap_decode(lpable)];
		}
		else {
			dp->flow_control = FLOW_CONTROL_NONE;
		}
		dp->mii_state = MII_STATE_MEDIA_SETUP;

		if (dp->ugc.usbgc_mii_an_delay > 0) {
			/*
			 * Some nics which detect media mode automatically
			 * take a little time to know the mode.
			 */
			dp->mii_timer = dp->ugc.usbgc_mii_an_delay;
			interval = drv_usectohz(20*1000);
			goto next;
		}
		goto next_nowait;

	case MII_STATE_MEDIA_SETUP:
		if (dp->mii_fixedmode || dp->ugc.usbgc_mii_an_oneshot) {
			/*
			 * write the result of auto negotiation back.
			 * XXX - This code didn't work on AN983B.
			 */
			val = usbgem_mii_read(dp, MII_CONTROL, &err);
			if (err != USB_SUCCESS) {
				goto usberr;
			}
			val &= ~(MII_CONTROL_SPEED | MII_CONTROL_FDUPLEX |
				 MII_CONTROL_ANE   | MII_CONTROL_RSAN);

			if (dp->full_duplex) {
				val |= MII_CONTROL_FDUPLEX;
			}

			switch (dp->speed) {
			case USBGEM_SPD_1000:
				val |= MII_CONTROL_1000MB;
				break;

			case USBGEM_SPD_100:
				val |= MII_CONTROL_100MB;
				break;

			default:
				/* for USBGEM_SPD_10, do nothing */
				break;
			}

			usbgem_mii_write(dp, MII_CONTROL, val, &err);
			if (err != USB_SUCCESS) {
				goto usberr;
			}
		}

		if (USBGEM_SET_MEDIA(dp) != USB_SUCCESS) {
			goto usberr;
		}

		if ((void *)(dp->ugc.usbgc_mii_tune_phy) != NULL) {
			USBGEM_MII_TUNE_PHY(dp, &err);
			if (err != USB_SUCCESS) {
				goto usberr;
			}
		}

		dp->mii_state = MII_STATE_LINKDOWN;
		dp->mii_timer = dp->ugc.usbgc_mii_linkdown_timeout;
		DPRINTF(2, (CE_CONT, "!%s: setup midia mode done", dp->name));
		dp->mii_supress_msg = B_FALSE;

		/* use short interval */
		interval = WATCH_INTERVAL_FAST;
		goto next_nowait;

	case MII_STATE_LINKDOWN:
		mii_stat = usbgem_mii_read(dp, MII_STATUS, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		if ((mii_stat & MII_STATUS_LINKUP) != 0) {
			uint16_t	mii_control;
			static char *fc_type[] = {
				"without",
				"with symmetric",
				"with tx",
				"with rx",
			};
			/*
			 * Link going up
			 */
			dp->mii_state = MII_STATE_LINKUP;
			dp->mii_supress_msg = B_FALSE;

			mii_control =usbgem_mii_read(dp, MII_CONTROL, &err);
			if (err != USB_SUCCESS) {
				goto usberr;
			}

			DPRINTF(0, (CE_CONT,
			"!%s: link up detected: mii_control:%b, mii_stat:%b",
				dp->name,
				mii_control, MII_CONTROL_BITS,
				mii_stat, MII_STATUS_BITS));

			/* restart tx watchdog */
			dp->tx_start_time = ddi_get_lbolt();

			/*
			 * MII_CONTROL_100MB and  MII_CONTROL_FDUPLEX are
			 * ignored when MII_CONTROL_ANE is set.
			 */
			cmn_err(CE_CONT,
			    "!%s: Link up: %d Mbps %s duplex %s flow control",
				dp->name,
				usbgem_speed_value[dp->speed],
				dp->full_duplex ? "full" : "half",
				fc_type[dp->flow_control]);

			interval = dp->ugc.usbgc_mii_link_watch_interval;
#ifdef S10
			gld_linkstate(dp->macinfo, GLD_LINKSTATE_UP);
#endif
			goto next;
		}

		dp->mii_supress_msg = B_TRUE;
		if (!dp->mii_fixedmode) {
			dp->mii_timer -= diff;
			if (dp->mii_timer <= 0) {
				/*
				 * the link down timer expired.
				 * need to restart auto-negotiation.
				 */
				linkdown_action =
				    dp->ugc.usbgc_mii_linkdown_timeout_action;
				goto restart_autonego;
			}
		}
		/* don't change mii_state */
		break;

	case MII_STATE_LINKUP:
		mii_stat = usbgem_mii_read(dp, MII_STATUS, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		if ((mii_stat & MII_STATUS_LINKUP) == 0) {
			/*
			 * Link going down
			 */
			cmn_err(CE_NOTE,
				"%s: link down detected: mii_stat:%b "
				"restarting auto-negotiation",
				dp->name, mii_stat, MII_STATUS_BITS);
#ifdef S10
			gld_linkstate(dp->macinfo, GLD_LINKSTATE_DOWN);
#endif
			if (!dp->mii_fixedmode) {
				/* need to restart auto-negotiation */
				linkdown_action =
					dp->ugc.usbgc_mii_linkdown_action;
				goto restart_autonego;
			}

			dp->mii_state = MII_STATE_LINKDOWN;
			dp->mii_timer = dp->ugc.usbgc_mii_linkdown_timeout;

			if ((void *)(dp->ugc.usbgc_mii_tune_phy) != NULL) {
				USBGEM_MII_TUNE_PHY(dp, &err);
				if (err != USB_SUCCESS) {
					goto usberr;
				}
			}
		}
		/* don't change mii_state */
		break;
	}
	interval = dp->ugc.usbgc_mii_link_watch_interval;
	goto next;

	/* Actions on the end of state routine */

restart_autonego:
	switch (linkdown_action) {
	case MII_ACTION_RESET:
		if (!dp->mii_supress_msg) {
			cmn_err(CE_CONT, "!%s: resetting PHY", dp->name);
		}
		dp->mii_supress_msg = B_TRUE;
		goto reset_phy;

	case MII_ACTION_NONE:
		dp->mii_supress_msg = B_TRUE;
		if (dp->ugc.usbgc_mii_an_oneshot) {
			goto autonego;
		}
		/* PHY will restart autonego automatically */
		dp->mii_state = MII_STATE_AUTONEGOTIATING;
		dp->mii_timer = dp->ugc.usbgc_mii_an_timeout;
		interval = dp->ugc.usbgc_mii_an_watch_interval;
		goto next;

	case MII_ACTION_RSA:
		if (!dp->mii_supress_msg) {
			cmn_err(CE_CONT, "!%s: restarting auto-negotiation",
				dp->name);
		}
		dp->mii_supress_msg = B_TRUE;
		use_rsan = B_TRUE;
		goto autonego;

	default:
		cmn_err(CE_PANIC, "!%s: unknowm linkdown action: %d",
			dp->name, dp->ugc.usbgc_mii_linkdown_action);
		dp->mii_supress_msg = B_TRUE;
	}
	/* NOTREACHED */

reset_phy:
#ifdef S10
	gld_linkstate(dp->macinfo, GLD_LINKSTATE_DOWN);
#endif
	if (!dp->mii_supress_msg) {
		cmn_err(CE_CONT, "!%s: resetting PHY", dp->name);
	}
	dp->mii_state = MII_STATE_RESETTING;
	dp->mii_timer = dp->ugc.usbgc_mii_reset_timeout;
	if (!dp->ugc.usbgc_mii_dont_reset) {
		usbgem_mii_write(dp, MII_CONTROL, MII_CONTROL_RESET, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
	}
	interval = WATCH_INTERVAL_FAST;
	goto next;

autonego:
#ifdef S10
	gld_linkstate(dp->macinfo, GLD_LINKSTATE_DOWN);
#endif
	if (!dp->mii_supress_msg) {
		cmn_err(CE_CONT, "!%s: auto-negotiation started", dp->name);
	}
	dp->mii_state = MII_STATE_AUTONEGOTIATING;
	dp->mii_timer = dp->ugc.usbgc_mii_an_timeout;

	val = usbgem_mii_read(dp, MII_CONTROL, &err);
	if (err != USB_SUCCESS) {
		goto usberr;
	}
	if (use_rsan || (val & MII_CONTROL_ANE) != 0) {
		/* restart auto nego */
		usbgem_mii_write(dp, MII_CONTROL,
				val | MII_CONTROL_ANE | MII_CONTROL_RSAN, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
	}
	else {
		/* enable auto nego */
		/* it doesn't work for mx98715 */
		usbgem_mii_write(dp, MII_CONTROL, val | MII_CONTROL_ANE, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
	}

	interval = dp->ugc.usbgc_mii_an_watch_interval;

next:
	return interval;

usberr:
	dp->mii_state = MII_STATE_UNKNOWN;
	return dp->ugc.usbgc_mii_link_watch_interval;
}

static void
usbgem_mii_link_watcher(struct usbgem_dev *dp)
{
	clock_t	next;

	mutex_enter(&dp->rxlock);
	next = usbgem_mii_link_check(dp);
	mutex_exit(&dp->rxlock);

	if (next != (clock_t) 0) {
		dp->link_watcher_id =
			timeout((void (*)(void *))& usbgem_mii_link_watcher,
				(void *)dp, next);
	}
}

int
usbgem_mii_init_default(struct usbgem_dev *dp)
{
	int		phy;
	uint16_t	status;
	uint16_t	xstatus;
	int		err;

	DPRINTF(3, (CE_CONT, "!%s: usbgem_mii_init: called", dp->name));

	/*
	 * Scan PHY
	 */
	dp->no_preamble = B_FALSE;
	USBGEM_MII_SYNC(dp, &err);
	if (err != USB_SUCCESS) {
		DPRINTF(0, (CE_CONT, "!%s: usbgem_mii_init: mii_sync failed",
			dp->name));
		goto usberr;
	}

	/* Try default phy first */
	if (dp->mii_phy_addr != 0) {
		status = usbgem_mii_read(dp, MII_STATUS, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		if (status != 0xffff && status != 0x0000) {
			goto PHY_found;
		}

		if (dp->mii_phy_addr < 0) {
			cmn_err(CE_NOTE,
			"!%s: failed to probe default internal/non-MII PHY",
				dp->name);
			return -1;
		}

		cmn_err(CE_NOTE,
			"!%s: failed to probe default MII PHY at %d",
			dp->name, dp->mii_phy_addr);
	}

	/* Try all possible address */
	for (phy = dp->ugc.usbgc_mii_addr_min; phy < 32; phy++) {
		dp->mii_phy_addr = phy;
		status = usbgem_mii_read(dp, MII_STATUS, &err);
		if (err != USB_SUCCESS) {
			DPRINTF(0, (CE_CONT,
				"!%s: usbgem_mii_init: mii_read(status) failed",
				dp->name));
			goto usberr;
		}

		if (status != 0xffff && status != 0x0000) {
			goto PHY_found;
		}
	}

	cmn_err(CE_NOTE, "!%s: no MII PHY found", dp->name);
	return USB_FAILURE;

PHY_found:
	dp->mii_phy_id  = usbgem_mii_read(dp, MII_PHYIDH, &err) << 16;
	if (err != USB_SUCCESS) {
		DPRINTF(0, (CE_CONT,
			"!%s: usbgem_mii_init: mii_read(PHYIDH) failed",
			dp->name));
		goto usberr;
	}
	dp->mii_phy_id |= usbgem_mii_read(dp, MII_PHYIDL, &err);
	if (err != USB_SUCCESS) {
		DPRINTF(0, (CE_CONT,
			"!%s: usbgem_mii_init: mii_read(PHYIDL) failed",
			dp->name));
		goto usberr;
	}

	dp->no_preamble = (status & MII_STATUS_MFPRMBLSUPR) != 0;

	if (dp->mii_phy_addr < 0) {
		cmn_err(CE_CONT, "!%s: using internal/non-MII PHY(0x%08x)",
			dp->name, dp->mii_phy_id);
	} else {
		cmn_err(CE_CONT, "!%s: MII PHY (0x%08x) found at %d",
			dp->name, dp->mii_phy_id, dp->mii_phy_addr);
	}

	cmn_err(CE_CONT, "!%s: PHY control:%b, status:%b, advert:%b, lpar:%b",
		dp->name,
		usbgem_mii_read(dp, MII_CONTROL, &err), MII_CONTROL_BITS,
		status, MII_STATUS_BITS,
		usbgem_mii_read(dp, MII_AN_ADVERT, &err), MII_ABILITY_BITS,
		usbgem_mii_read(dp, MII_AN_LPABLE, &err), MII_ABILITY_BITS);
#ifdef USBGEM_GIGA
	cmn_err(CE_CONT, "!%s: xstatus:%b",
		dp->name,
		usbgem_mii_read(dp, MII_XSTATUS, &err), MII_XSTATUS_BITS);
#endif
	if (!dp->mii_fixedmode && (status & MII_STATUS_CANAUTONEG) == 0) {
		dp->mii_fixedmode = B_TRUE;
		/* fix speed and duplex mode. check half duplex first */
		if ((status & MII_STATUS_10) != 0) {
			dp->speed	= USBGEM_SPD_10;
			dp->full_duplex = B_FALSE;
		}
		else if ((status & MII_STATUS_100_BASEX) != 0) {
			dp->speed	= USBGEM_SPD_100;
			dp->full_duplex = B_FALSE;
		}
		else if ((status & MII_STATUS_10_FD) != 0) {
			dp->speed	= USBGEM_SPD_10;
			dp->full_duplex = B_TRUE;
		}
		else if ((status & MII_STATUS_100_BASEX_FD) != 0) {
			dp->speed	= USBGEM_SPD_100;
			dp->full_duplex = B_TRUE;
		}
		else if ((status & MII_STATUS_100_BASE_T4) != 0) {
			dp->speed	= USBGEM_SPD_100;
			dp->full_duplex = B_FALSE;
		}
	}

	return USB_SUCCESS;

usberr:
	return USB_FAILURE;
}

static int
usbgem_mii_start(struct usbgem_dev *dp)
{
	int		phy;
	uint16_t	adv;
	uint16_t	status;
	uint16_t	val;
	clock_t		interval;
	int		err;

#ifdef S10
	gld_linkstate(dp->macinfo, GLD_LINKSTATE_DOWN);
#endif
	if (dp->ugc.usbgc_mii_dont_reset) {
		/* avoid resetting PHY */
		if (USBGEM_MII_CONFIG(dp, &err) != 0) {
			cmn_err(CE_WARN, "!%s: usbgem_mii_config failed",
				dp->name);
		}
		if (dp->mii_fixedmode) {
			dp->mii_state = MII_STATE_LINKDOWN;
			dp->mii_timer = dp->ugc.usbgc_mii_linkdown_timeout;
			goto x;
		}
		dp->mii_state = MII_STATE_AUTONEGOTIATING;
		dp->mii_timer = dp->ugc.usbgc_mii_an_timeout;
		val = usbgem_mii_read(dp, MII_CONTROL, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		usbgem_mii_write(dp, MII_CONTROL,
			val | MII_CONTROL_ANE | MII_CONTROL_RSAN, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		interval = dp->ugc.usbgc_mii_an_watch_interval;
	}
	else {
		/* reset PHY */
		dp->mii_state = MII_STATE_RESETTING;
		dp->mii_timer = dp->ugc.usbgc_mii_reset_timeout;
		usbgem_mii_write(dp, MII_CONTROL, MII_CONTROL_RESET, &err);
		if (err != USB_SUCCESS) {
			goto usberr;
		}
		interval = WATCH_INTERVAL_FAST;
	}
x:
	/* schedule first call of gem_link_watcher */
	dp->mii_last_check = ddi_get_lbolt();
	dp->link_watcher_id =
		timeout((void (*)(void *))& usbgem_mii_link_watcher,
			(void *)dp, interval);

	return USB_SUCCESS;

usberr:
	return USB_FAILURE;
}

static void
usbgem_mii_stop(struct usbgem_dev *dp)
{
	timeout_id_t	old_id;

	/* Ensure timer routine stopped */
	if (dp->link_watcher_id != 0) {
		do {
			untimeout(old_id = dp->link_watcher_id);
		} while (old_id != dp->link_watcher_id);
		dp->link_watcher_id = 0;
	}
}

void
usbgem_generate_macaddr(struct usbgem_dev *dp, uint8_t *mac)
{
	int		i;
	uint_t		val;

	cmn_err(CE_CONT,
	"!%s: using a local mac address, do not use this for long time",
		dp->name);
	val = ddi_get_lbolt() ^ ddi_get_time();

	/* generate a local mac address */
	mac[0] = 0x02;
	mac[1] = 0x00;
	mac[2] = val >> 24;
	mac[3] = val >> 16;
	mac[4] = val >> 8;
	mac[5] = val;
}

boolean_t
usbgem_get_mac_addr_conf(struct usbgem_dev *dp)	/* done */
{
	char		propname[32];
	char		valstr[32];
	uint8_t		mac[ETHERADDRL];
	char		*cp;
	int		len;
	int		c;
	int		i;
	int		j;
	int		v;
	int		d;
	int		ored;

	/*
	 * Get ethernet address from .conf file
	 */
	valstr[0] = 0;
	sprintf(propname, "%s-mac-addr", dp->name);
	len = sizeof(valstr);
	if ((ddi_getlongprop_buf(DDI_DEV_T_ANY, dp->dip,
			DDI_PROP_DONTPASS, propname, valstr,
			&len)) != DDI_PROP_SUCCESS) {
#ifdef notdef
		cmn_err(CE_CONT,
			"!%s: trying to read %s property from .conf: undefined",
			dp->name, propname);
#endif
		return B_FALSE;
	}

	if (strlen(valstr) != ETHERADDRL*3-1) {
		goto syntax_err;
	}

	cp = valstr;
	j  = 0;
	ored = 0;
	while (1) {
		v = 0;
		for (i = 0; i < 2; i++) {
			c = *cp++;

			if (c >= 'a' && c <= 'f') {
				d = c - 'a' + 10;
			}
			else if (c >= 'A' && c <= 'F') {
				d = c - 'A' + 10;
			}
			else if (c >= '0' && c <= '9') {
				d = c - '0';
			}
			else {
				goto syntax_err;
			}
			v = (v << 4) | d;
		}

		mac[j++] = v;
		ored |= v;
		if (j == ETHERADDRL) {
			/* done */
			break;
		}

		c = *cp++;
		if (c != ':') {
			goto syntax_err;
		}
	}

	if (ored == 0) {
		usbgem_generate_macaddr(dp, mac);
	}

	for (i = 0; i < ETHERADDRL; i++) {
		dp->dev_addr.ether_addr_octet[i] = mac[i];
	}
	return B_TRUE;

syntax_err:
	cmn_err(CE_CONT,
		"!%s: read mac addr: trying .conf: syntax err %s",
		dp->name, valstr);

	return B_FALSE;
}

/* ======================================================================== */
/*
 * attach/detatch support
 */
/* ======================================================================== */
static void
usbgem_read_conf(struct usbgem_dev *dp)			/* done */
{
	char			propname[32];
	char			*valstr;
	int			*valp;
	uint_t			len;

#define GET_PROP_INT(dp, prop, x) {	\
	sprintf(propname, prop, (dp)->name);	\
	if (ddi_prop_lookup_int_array(DDI_DEV_T_ANY, (dp)->dip,	\
			DDI_PROP_DONTPASS, propname, &valp,	\
			&len) == DDI_PROP_SUCCESS) {	\
		if (len == 1) {	\
			x = *valp;	\
		}	\
		ddi_prop_free(valp);	\
	}	\
}

#define GET_PROP_BOOL(dp, prop, x) {	\
	sprintf(propname, prop, (dp)->name);	\
	if (ddi_prop_lookup_int_array(DDI_DEV_T_ANY, (dp)->dip,	\
			DDI_PROP_DONTPASS, propname, &valp,	\
			&len) == DDI_PROP_SUCCESS) {	\
		if (len == 1) {	\
			x = (*valp != 0);	\
		}	\
		ddi_prop_free(valp);	\
	}	\
}

	/*
	 * Get media mode infomation from .conf file
	 */
	sprintf(propname, "%s-duplex", dp->name);
	if ((ddi_prop_lookup_string(DDI_DEV_T_ANY, dp->dip,
			DDI_PROP_DONTPASS, propname, &valstr
			)) == DDI_PROP_SUCCESS) {
		dp->mii_fixedmode = B_TRUE;
		if (strcmp(valstr, "full") == 0) {
			dp->full_duplex = B_TRUE;
		}
		else if (strcmp(valstr, "half") == 0) {
			dp->full_duplex = B_FALSE;
		}
		else {
			cmn_err(CE_WARN,
				"!%s: property %s: illegal value (%s)",
				dp->name, propname, valstr);
		}
		ddi_prop_free(valstr);
	}

	sprintf(propname, "%s-speed", dp->name);
	if (ddi_prop_lookup_int_array(DDI_DEV_T_ANY, dp->dip,
				DDI_PROP_DONTPASS, propname, &valp,
				&len) == DDI_PROP_SUCCESS) {
		dp->mii_fixedmode = B_TRUE;
		if (*valp == 1000 && len == 1) {
			dp->speed = USBGEM_SPD_1000;
		}
		else if (*valp == 100 && len == 1) {
			dp->speed = USBGEM_SPD_100;
		}
		else if (*valp == 10 && len == 1) {
			dp->speed = USBGEM_SPD_10;
		}
		else {
			cmn_err(CE_WARN,
				"!%s: property %s: illegal value (%d)",
				dp->name, propname, *valp);
		}
		ddi_prop_free(valp);
	}

	sprintf(propname, "%s-flow-control", dp->name);
	if (ddi_prop_lookup_int_array(DDI_DEV_T_ANY, dp->dip,
			DDI_PROP_DONTPASS, propname, &valp,
			&len) == DDI_PROP_SUCCESS) {

		if (*valp > FLOW_CONTROL_RX_PAUSE ||
		    *valp < FLOW_CONTROL_NONE || len != 1) {
			cmn_err(CE_WARN,
				"!%s: property %s: illegal value:%d, len:%d",
				dp->name, propname, *valp, len);
		}
		else {
			dp->ugc.usbgc_flow_control = *valp;
		}
		ddi_prop_free(valp);
	}

	if (dp->ugc.usbgc_mii_mode == USBGEM_MODE_1000BASET) {
		dp->anadv_1000fdx = B_TRUE;
		GET_PROP_BOOL(dp, "%s-anadv-1000fdx", dp->anadv_1000fdx);
		dp->anadv_1000hdx = B_TRUE;
		GET_PROP_BOOL(dp, "%s-anadv-1000hdx", dp->anadv_1000hdx);
	}

	dp->anadv_100t4 = B_TRUE;
	GET_PROP_BOOL(dp, "%s-anadv-100t4", dp->anadv_100t4);
	dp->anadv_100fdx = B_TRUE;
	GET_PROP_BOOL(dp, "%s-anadv-100fdx", dp->anadv_100fdx);
	dp->anadv_100hdx = B_TRUE;
	GET_PROP_BOOL(dp, "%s-anadv-100hdx", dp->anadv_100hdx);
	dp->anadv_10fdx = B_TRUE;
	GET_PROP_BOOL(dp, "%s-anadv-10fdx", dp->anadv_10fdx);
	dp->anadv_10hdx = B_TRUE;
	GET_PROP_BOOL(dp, "%s-anadv-10hdx", dp->anadv_10hdx);

	sprintf(propname, "%s-nointr", dp->name);
	if (ddi_prop_lookup_int_array(DDI_DEV_T_ANY, dp->dip,
			DDI_PROP_DONTPASS, propname, &valp,
			&len) == DDI_PROP_SUCCESS) {
		if (len == 1) {
			dp->misc_flag &= ~USBGEM_NOINTR;
			if (*valp != 0) {
				dp->misc_flag |= USBGEM_NOINTR;
				cmn_err(CE_CONT,
					"!%s: polling mode enabled", dp->name);
			}
		}
		ddi_prop_free(valp);
	}

	GET_PROP_INT(dp, "%s-mtuthr", dp->mtu);
	GET_PROP_INT(dp, "%s-txthr", dp->txthr);
	GET_PROP_INT(dp, "%s-txmaxdma", dp->txmaxdma);
	GET_PROP_INT(dp, "%s-rxthr", dp->rxthr);
	GET_PROP_INT(dp, "%s-rxmaxdma", dp->rxmaxdma);

#undef GET_PROP_INT
#undef GET_PROP_BOOL
}

#ifdef CTRL_TIMEOUT
static void
usbgem_ctrl_cb(usb_pipe_handle_t ph, usb_ctrl_req_t *creq)
{
	struct usbgem_dev	*dp;
	boolean_t		send;
	int			timeout;
	int			err;
	usb_flags_t		flags;
	mblk_t			*newmp;
	mblk_t			*mp;

	dp = (struct usbgem_dev *)creq->ctrl_client_private;
	mp = creq->ctrl_data;

	DPRINTF(0, (CE_CONT, "!%s: usbgem_ctrl_cb:%s(%d) cb_flags:0x%x",
		dp->name,
		usb_str_cr(creq->ctrl_completion_reason),
		creq->ctrl_completion_reason,
		creq->ctrl_cb_flags));
#ifdef notdef
	if (req->bulk_completion_reason == 12) {
		/* time out */
		usb_pipe_reset(dp->dip, dp->bulkout_pipe, 0, NULL, 0);
	}
#endif
}

int
usbgem_ctrl_out(struct usbgem_dev *dp,
	uint8_t reqt, uint8_t req, uint16_t val, uint16_t ix, uint16_t len,
	void *bp, int size)
{
	usb_ctrl_req_t		*creq;
	usb_flags_t		flags;
	mblk_t			*data;
	int			i;
	int			ret;

	DPRINTF(0, (CE_CONT,
"!%s: usbgem_ctrl_out reqt:0x%02x req:0x%02x val:0x%04x ix:0x%04x len:0x%02x "
		"bp:0x%p nic_state:%d",
		dp->name, reqt, req, val, ix, len, bp, dp->nic_state));

	ret  = USB_SUCCESS;

	if (dp->nic_state == NIC_DISCONNECTED) {
		return USB_PIPE_ERROR;
	}

	creq = usb_alloc_ctrl_req(dp->dip, max(len, size), USB_FLAGS_SLEEP);
	if (creq == NULL) {
		cmn_err(CE_WARN,
		"!%s: usbgem_ctrl_out: failed to allocate ctrl req",
			dp->name);
		return USB_FAILURE;
	}

	data = creq->ctrl_data;
	if (size > 0) {
		bcopy(bp, data->b_rptr, size);
		data->b_wptr = data->b_rptr + size;
	}

	for (i = usbgem_ctrl_retry; i > 0; i--) {
		creq->ctrl_bmRequestType = reqt;
		creq->ctrl_bRequest	= req;
		creq->ctrl_wValue	= val;
		creq->ctrl_wIndex	= ix;
		creq->ctrl_wLength	= len;
		creq->ctrl_data		= data;
		creq->ctrl_timeout	= 0;
		creq->ctrl_client_private = (usb_opaque_t) dp;
		creq->ctrl_attributes	= (len > size)
						? USB_ATTRS_SHORT_XFER_OK : 0;
		creq->ctrl_cb		= NULL;
		creq->ctrl_exc_cb	= NULL;
		creq->ctrl_completion_reason = 0;
		creq->ctrl_cb_flags	= 0; 
		flags = USB_FLAGS_SLEEP;

		ret = usb_pipe_ctrl_xfer(DEFAULT_PIPE(dp), creq, flags);

		if (ret == USB_SUCCESS) {
			break;
		}

		cmn_err(CE_WARN,
		"!%s: usb_pipe_ctrl_xfer failed: "
		"reqt:0x%x req:0x%x val:0x%x ix:0x%x len:0x%x "
		"ret:%d cr:(%d), cb_flags:0x%x %s",
			dp->name, reqt, req, val, ix, len,
			ret, /*usb_str_cr(creq->ctrl_completion_reason),*/
			creq->ctrl_completion_reason,
			creq->ctrl_cb_flags,
			(i > 1) ? "retrying..." : "fatal");

	}

err:
#ifdef notdef
	if (data != NULL) {
		freemsg(data);
	}
#endif
	if (ret != USB_SUCCESS) {
		/* we detected a fatal error */
		dp->nic_state = NIC_ERROR;
	}

	usb_free_ctrl_req(creq);

	return ret;
}

int
usbgem_ctrl_in(struct usbgem_dev *dp,
	uint8_t reqt, uint8_t req, uint16_t val, uint16_t ix, uint16_t len,
	void *bp, int size)
{
	usb_ctrl_req_t		*creq;
	usb_flags_t		flags;
	mblk_t			*data;
	int			i;
	int			ret;

	DPRINTF(4, (CE_CONT, "!%s: usbgem_ctrl_in "
		"reqt:0x%02x req:0x%02x val:0x%04x ix:0x%04x len:0x%02x "
		"bp:0x%p nic_state:%d",
		dp->name, reqt, req, val, ix, len, bp, dp->nic_state));

	data = NULL;
	ret  = USB_SUCCESS;

	if (dp->nic_state == NIC_DISCONNECTED) {
		return USB_PIPE_ERROR;
	}

	creq = usb_alloc_ctrl_req(dp->dip, 0, USB_FLAGS_SLEEP);
	if (creq == NULL) {
		cmn_err(CE_WARN,
		"!%s: usbgem_ctrl_in: failed to allocate ctrl req",
			dp->name);
		return USB_FAILURE;
	}

	if (size > 0) {
		if ((data = allocb(size, 0)) == NULL) {
			goto err;
		}
#ifdef notdef
		bcopy(bp, data->b_rptr, size);
#endif
		data->b_wptr = data->b_rptr + size;
	}

	for (i = usbgem_ctrl_retry; i > 0; i--) {
		creq->ctrl_bmRequestType = reqt;
		creq->ctrl_bRequest	= req;
		creq->ctrl_wValue	= val;
		creq->ctrl_wIndex	= ix;
		creq->ctrl_wLength	= len;
		creq->ctrl_data		= data;
		creq->ctrl_timeout	= 10;
		creq->ctrl_client_private = (usb_opaque_t) dp;
		creq->ctrl_attributes	= 0;
		creq->ctrl_cb		= usbgem_ctrl_cb;
		creq->ctrl_exc_cb	= usbgem_ctrl_cb;
		creq->ctrl_completion_reason = 0;
		creq->ctrl_cb_flags	= 0; 
		flags = USB_FLAGS_SLEEP;

		ret = usb_pipe_ctrl_xfer(DEFAULT_PIPE(dp), creq, flags);

		if (ret == USB_SUCCESS) {
			break;
		}

		cmn_err(CE_WARN,
		"!%s: usb_pipe_ctrl_xfer failed: "
		"reqt:0x%x req:0x%x val:0x%x ix:0x%x len:0x%x "
		"ret:%d cr:%s(%d), cb_flags:0x%x %s",
			dp->name, reqt, req, val, ix, len,
			ret, usb_str_cr(creq->ctrl_completion_reason),
			creq->ctrl_completion_reason,
			creq->ctrl_cb_flags,
			(i > 1) ? "retrying..." : "fatal");
	}

	bcopy(data->b_rptr, bp, size);
err:
	if (data != NULL) {
		freemsg(data);
	}

	if (ret != USB_SUCCESS) {
		/* we detected a fatal error */
		dp->nic_state = NIC_ERROR;
	}

	usb_free_ctrl_req(creq);

	return ret;
}

#else

int
usbgem_ctrl_out(struct usbgem_dev *dp,
	uint8_t reqt, uint8_t req, uint16_t val, uint16_t ix, uint16_t len,
	void *bp, int size)
{
	mblk_t			*data;
	usb_ctrl_setup_t	setup;
	usb_cr_t		completion_reason;
        usb_cb_flags_t		cb_flags;
        usb_flags_t		flags;
	int			i;
	int			ret;

	DPRINTF(4, (CE_CONT, "!%s: usbgem_ctrl_out "
		"reqt:0x%02x req:0x%02x val:0x%04x ix:0x%04x len:0x%02x "
		"bp:0x%p nic_state:%d",
		dp->name, reqt, req, val, ix, len, bp, dp->nic_state));

	if (dp->nic_state == NIC_DISCONNECTED) {
		return USB_PIPE_ERROR;
	}

	if (size > 0) {
		if ((data = allocb(size, 0)) == NULL) {
			return USB_FAILURE;
		}

		bcopy(bp, data->b_rptr, size);
		data->b_wptr = data->b_rptr + size;
	}
	else {
		data = NULL;
	}

	setup.bmRequestType = reqt;
	setup.bRequest = req;
	setup.wValue   = val;
	setup.wIndex   = ix;
	setup.wLength  = len;
	setup.attrs    = 0;	/* attributes */

	for (i = usbgem_ctrl_retry; i > 0; i--) {
		completion_reason = 0;
		cb_flags = 0;

		ret = usb_pipe_ctrl_xfer_wait(DEFAULT_PIPE(dp), &setup, &data,
			&completion_reason, &cb_flags, 0);

		if (ret == USB_SUCCESS) {
			break;
		}
		if (i == 1) {
			cmn_err(CE_WARN,
			"!%s: usbgem_ctrl_out failed: "
			"reqt:0x%x req:0x%x val:0x%x ix:0x%x len:0x%x "
			"ret:%d cr:%s(%d), cb_flags:0x%x %s",
				dp->name, reqt, req, val, ix, len,
				ret, usb_str_cr(completion_reason),
				completion_reason,
				cb_flags,
				(i > 1) ? "retrying..." : "fatal");
		}
	}

	if (data != NULL) {
		freemsg(data);
	}

	if (ret != USB_SUCCESS) {
		/* we detected a fatal error */
		dp->nic_state = NIC_ERROR;
	}

	return ret;
}

int
usbgem_ctrl_in(struct usbgem_dev *dp, 
	uint8_t reqt, uint8_t req, uint16_t val, uint16_t ix, uint16_t len,
	void *bp, int size)
{
	mblk_t			*data;
	usb_ctrl_setup_t	setup;
	usb_cr_t		completion_reason;
        usb_cb_flags_t		cb_flags;
	int			i;
	int			ret;
	int			reclen;

	DPRINTF(4, (CE_CONT, "!%s: usbgem_ctrl_in 0x%x"
		"reqt:0x%02x req:0x%02x val:0x%04x ix:0x%04x len:0x%02x "
		"bp:x%p nic_state:%d",
		dp->name, reqt, req, val, ix, len, bp, dp->nic_state));

	if (dp->nic_state == NIC_DISCONNECTED) {
		return USB_PIPE_ERROR;
	}

	data = NULL;

	setup.bmRequestType = reqt;
	setup.bRequest = req;
	setup.wValue   = val;
	setup.wIndex   = ix;
	setup.wLength  = len;
	setup.attrs    = USB_ATTRS_AUTOCLEARING;	/* XXX */

	for (i = usbgem_ctrl_retry; i > 0; i--) {
		completion_reason = 0;
		cb_flags = 0;
		ret = usb_pipe_ctrl_xfer_wait(DEFAULT_PIPE(dp), &setup, &data,
			&completion_reason, &cb_flags, 0);

		if (ret == USB_SUCCESS) {
			reclen = msgdsize(data);
			bcopy(data->b_rptr, bp, min(reclen, size));
			break;
		}
		if (i == 1) {
			cmn_err(CE_WARN,
			"%s: usbgem_ctrl_in failed: "
			"reqt:0x%x req:0x%x val:0x%x ix:0x%x len:0x%x "
			"ret:%d cr:%s(%d) cb_flags:0x%x %s",
				dp->name, reqt, req, val, ix, len,
				ret,
				usb_str_cr(completion_reason),
				completion_reason,
				cb_flags,
				(i > 1) ? "retrying..." : "fatal");
		}
	}

	if (data) {
		freemsg(data);
	}

	if (ret != USB_SUCCESS) {
		/* we detected a fatal error */
		dp->nic_state = NIC_ERROR;
	}

	return ret;
}
#endif /* CTRL_TIMEOUT */

int
usbgem_ctrl_out_val(struct usbgem_dev *dp, 
	uint8_t reqt, uint8_t req, uint16_t val, uint16_t ix, uint16_t len,
	uint32_t v)
{
	uint8_t	buf[4];

	/* convert to little endian from native byte order */
	switch (len) {
	case 4:
		buf[3] = v >> 24;
		buf[2] = v >> 16;
		/* fall down */
	case 2:
		buf[1] = v >> 8;
		/* fall down */
	case 1:
		buf[0] = v;
	}

	return usbgem_ctrl_out(dp, reqt, req, val, ix, len, buf, len);
}

int
usbgem_ctrl_in_val(struct usbgem_dev *dp, 
	uint8_t reqt, uint8_t req, uint16_t val, uint16_t ix, uint16_t len,
	void *valp)
{
	uint8_t		buf[4];
	uint_t		v;
	int		err;

#ifdef SANITY
	bzero(buf, sizeof(buf));
#endif
	err = usbgem_ctrl_in(dp, reqt, req, val, ix, len, buf, len);
	if (err == USB_SUCCESS) {
		v = 0;
		switch (len) {
		case 4:
			v |= buf[3] << 24;
			v |= buf[2] << 16;
			/* fall down */
		case 2:
			v |= buf[1] << 8;
			/* fall down */
		case 1:
			v |= buf[0];
		}

		switch (len) {
		case 4:
			*(uint32_t *)valp = v;
			break;
		case 2:
			*(uint16_t *)valp = v;
			break;
		case 1:
			*(uint8_t *)valp = v;
			break;
		}
	}
	return err;
}

/*
 * Attach / detach / disconnect / reconnect management
 */
static int
usbgem_open_pipes(struct usbgem_dev  *dp)
{
	int			i;
	int			ret;
	int			ifnum;
	int			alt;
	usb_client_dev_data_t	*reg_data;
	usb_ep_data_t		*ep_tree_node;

	DPRINTF(1, (CE_CONT, "!%s: usbgem_open_pipes: called", dp->name));
#ifdef notdef
	ifnum = usb_get_if_number(dp->dip);
#else
	ifnum = dp->ugc.usbgc_ifnum;
	alt   = dp->ugc.usbgc_alt;
#endif
	ep_tree_node = usb_get_ep_data(dp->dip, dp->reg_data, ifnum, alt,
				USB_EP_ATTR_BULK, USB_EP_DIR_IN);
	if (ep_tree_node == NULL) {
		cmn_err(CE_WARN, "!%s: usbgem_open_pipes: ep_bulkin is NULL",
			dp->name);
		goto err;
	}
	dp->ep_bulkin = &ep_tree_node->ep_descr;

	ep_tree_node = usb_get_ep_data(dp->dip, dp->reg_data,
				ifnum, alt,
				USB_EP_ATTR_BULK, USB_EP_DIR_OUT);
	if (ep_tree_node == NULL) {
		cmn_err(CE_WARN, "!%s: usbgem_open_pipes: ep_bulkout is NULL",
			dp->name);
		goto err;
	}
	dp->ep_bulkout = &ep_tree_node->ep_descr;

	ep_tree_node = usb_get_ep_data(dp->dip, dp->reg_data, ifnum, alt,
				USB_EP_ATTR_INTR, USB_EP_DIR_IN);
	if (ep_tree_node == NULL) {
		/* don't care */
		cmn_err(CE_CONT, "%s: ep_intr is NULL", dp->name);
	}
	dp->ep_intr = &ep_tree_node->ep_descr;

	/* XXX -- no need to open default pipe */

	/* open bulk out pipe */
	bzero(&dp->policy_bulkout, sizeof(usb_pipe_policy_t));
	dp->policy_bulkout.pp_max_async_reqs = dp->ugc.usbgc_tx_list_max;
	if ((ret = usb_pipe_open(dp->dip,
			dp->ep_bulkout, &dp->policy_bulkout, USB_FLAGS_SLEEP,
			&dp->bulkout_pipe)) != USB_SUCCESS) {

		cmn_err(CE_WARN,
		"%s: usbgem_open_pipes: err:%x: failed to open bulk-out pipe",
			dp->name, ret);

		goto err;
	}
	DPRINTF(1, (CE_CONT, "!%s: bulkout_pipe opened", dp->name));

	/* open bulk in pipe */
	bzero(&dp->policy_bulkin, sizeof(usb_pipe_policy_t));
	dp->policy_bulkin.pp_max_async_reqs = 100; /* XXX */
	if ((ret = usb_pipe_open(dp->dip,
			dp->ep_bulkin, &dp->policy_bulkin, USB_FLAGS_SLEEP,
			&dp->bulkin_pipe)) != USB_SUCCESS) {

		cmn_err(CE_WARN,
		"%s: usbgem_open_pipes: ret:%x failed to open bulk-in pipe",
			dp->name, ret);

		goto err;
	}
	DPRINTF(1, (CE_CONT, "!%s: bulkin_pipe opened", dp->name));

	if (dp->ep_intr != NULL) {
		/* open interrupt pipe */
		bzero(&dp->policy_interrupt, sizeof(usb_pipe_policy_t));
		dp->policy_interrupt.pp_max_async_reqs = 1; /* XXX */
		if ((ret = usb_pipe_open(dp->dip, dp->ep_intr,
				&dp->policy_interrupt, USB_FLAGS_SLEEP,
				&dp->intr_pipe)) != USB_SUCCESS) {

			cmn_err(CE_WARN,
				"%s: usbgem_open_pipes: "
				"ret:%x failed to open interrupt pipe",
				dp->name, ret);

			goto err;
		}
	}
	DPRINTF(1, (CE_CONT, "!%s: intr_pipe opened", dp->name));

	return USB_SUCCESS;

err:
	if (dp->bulkin_pipe != NULL) {
		usb_pipe_close(dp->dip,
			dp->bulkin_pipe, USB_FLAGS_SLEEP, NULL, 0);
		dp->bulkin_pipe = NULL;
	}
	if (dp->bulkout_pipe != NULL) {
		usb_pipe_close(dp->dip,
			dp->bulkout_pipe, USB_FLAGS_SLEEP, NULL, 0);
		dp->bulkout_pipe = NULL;
	}
	if (dp->intr_pipe != NULL) {
		usb_pipe_close(dp->dip,
			dp->intr_pipe, USB_FLAGS_SLEEP, NULL, 0);
		dp->intr_pipe = NULL;
	}

	return USB_FAILURE;
}

static int
usbgem_close_pipes(struct usbgem_dev  *dp)
{
	if (dp->intr_pipe != NULL) {
		(void) usb_pipe_close(dp->dip, dp->intr_pipe,
					USB_FLAGS_SLEEP, NULL, 0);
		dp->intr_pipe = NULL;
	}

	ASSERT(dp->bulkin_pipe);
	usb_pipe_close(dp->dip, dp->bulkin_pipe, USB_FLAGS_SLEEP, NULL, 0);
	dp->bulkin_pipe = NULL;

	ASSERT(dp->bulkout_pipe);
	usb_pipe_close(dp->dip, dp->bulkout_pipe, USB_FLAGS_SLEEP, NULL, 0);
	dp->bulkout_pipe = NULL;

	return USB_SUCCESS;
}

static int
usbgem_disconnect_cb(dev_info_t *dip)
{
	gld_mac_info_t		*macinfo;
	struct usbgem_dev	*dp;

	macinfo = (gld_mac_info_t *)ddi_get_driver_private(dip);
	dp = (struct usbgem_dev *)macinfo->gldm_private;

	cmn_err(CE_WARN, "%s: the usb device was disconnected", dp->name);
#ifdef notdef
	/* stop mii watcher */
	usbgem_mii_stop(dp);
#endif
	/* stop nic activity */
	mutex_enter(&dp->rxlock);
	mutex_enter(&dp->txlock);
	dp->nic_state = NIC_DISCONNECTED;
	mutex_exit(&dp->txlock);
	mutex_exit(&dp->rxlock);

	/* cancel all of pended packets */
	usb_pipe_reset(dp->dip, dp->bulkin_pipe, USB_FLAGS_SLEEP, NULL, 0);
	usb_pipe_reset(dp->dip, dp->bulkout_pipe, USB_FLAGS_SLEEP, NULL, 0);

        DPRINTF(0, (CE_CONT, "!%s: usbgem_disconnect: rx_line:%d",
		dp->name, dp->rx_line_busy));

	/* ensure calback routines have been executed for canceled packets */
	mutex_enter(&dp->rxlock);
	while (dp->rx_line_busy > 0) {
		cv_wait(&dp->drain_cv, &dp->rxlock);
	}
	mutex_exit(&dp->rxlock);
	mutex_enter(&dp->txlock);
	while (dp->tx_tail - dp->tx_head > 0) {
		cv_wait(&dp->drain_cv, &dp->txlock);
	}
	mutex_exit(&dp->txlock);

	/*
	 * Here we free all memory resource allocated, becase we
	 * cannot free them while the usb device is disconnected.
	 */
	(void)usbgem_free_memory(dp);

        return USB_SUCCESS;
}

static int
usbgem_reconnect_cb(dev_info_t *dip)
{
	gld_mac_info_t		*macinfo;
	struct usbgem_dev	*dp;
	int			err;

	macinfo = (gld_mac_info_t *)ddi_get_driver_private(dip);
	dp = (struct usbgem_dev *)macinfo->gldm_private;

	cmn_err(CE_WARN, "%s: the usb device was reconnected", dp->name);

	mutex_enter(&dp->rxlock);
	mutex_enter(&dp->txlock);

	usbgem_close_pipes(dp);
	usbgem_open_pipes(dp);

	/* tell tx timeout watcher to restart the device */
	dp->nic_state = NIC_ERROR;

	mutex_exit(&dp->rxlock);
	mutex_exit(&dp->txlock);

	/* allocate memory resources again */
	err = usbgem_alloc_memory(dp);

	/* reset mii and start mii link watcher */
#ifdef notdef
	USBGEM_RESET_CHIP(dp);
#endif
	USBGEM_MII_INIT(dp);
#ifdef notdef
	usbgem_mii_start(dp);
#endif
x:
	return err;
}

static int
usbgem_pre_suspend_event_cb(dev_info_t *dip)
{
	gld_mac_info_t		*macinfo;
	struct usbgem_dev	*dp;

	macinfo = (gld_mac_info_t *)ddi_get_driver_private(dip);
	dp = (struct usbgem_dev *)macinfo->gldm_private;

	DPRINTF(0, (CE_CONT, "!%s: pre_suspend: enter", dp->name));

	/* not implemented yet */

	return USB_SUCCESS;
}

static int
usbgem_post_resume_event_callback(dev_info_t *dip)
{
	gld_mac_info_t		*macinfo;
	struct usbgem_dev	*dp;

	macinfo = (gld_mac_info_t *)ddi_get_driver_private(dip);
	dp = (struct usbgem_dev *)macinfo->gldm_private;

	DPRINTF(0, (CE_CONT, "!%s: post_resume: enter", dp->name));

	/* not implemented yet */

	return USB_SUCCESS;
}

#define USBGEM_LOCAL_DATA_SIZE(gc)	\
	(sizeof(struct usbgem_dev) + USBGEM_MCALLOC +	\
	sizeof(struct txbuf)*((gc)->usbgc_tx_list_max))

struct usbgem_dev *
usbgem_do_attach(dev_info_t *dip,
	struct usbgem_conf *gc, void *lp, int lmsize)
{
	struct usbgem_dev	*dp;
	int			i;
	gld_mac_info_t		*macinfo;
	int			ret;
	int			unit;
	int			err;

	unit = ddi_get_instance(dip);

	DPRINTF(2, (CE_CONT, "!usbgem%d: usbgem_do_attach: called cmd:ATTACH",
		unit));

	macinfo = gld_mac_alloc(dip);
	if (macinfo == NULL) {
		return NULL;
	}

	/*
	 * Allocate soft data structure
	 */
	dp = (struct usbgem_dev *)
		kmem_zalloc(USBGEM_LOCAL_DATA_SIZE(gc), KM_SLEEP);
	if (dp == NULL) {
		gld_mac_free(macinfo);
		return NULL;
	}
	/* link to private area */
	dp->private   = lp;
	dp->priv_size = lmsize;

	dp->mc_count     = 0;
	dp->mc_count_req = 0;
	dp->mc_list = (struct mcast_addr *) &dp[1];

	dp->dip = dip;
	dp->macinfo = macinfo;
	bcopy(gc->usbgc_name, dp->name, USBGEM_NAME_LEN);
#ifdef USB10
	if (usb_register_client(dip, USBDRV_VERSION,
			&dp->reg_data, USB_PARSE_LVL_ALL, 0) !=
		USB_SUCCESS) {
		cmn_err(CE_WARN,
			"%s: %s: usb_register_client failed",
			dp->name, __func__);
		goto err_free_private;
	}
#else
	if (usb_client_attach(dip, USBDRV_VERSION, 0) != USB_SUCCESS) {
		cmn_err(CE_WARN,
			"%s: %s: usb_client_attach failed",
			dp->name, __func__);
		goto err_free_private;
	}

	if (usb_get_dev_data(dip, &dp->reg_data,
			USB_PARSE_LVL_ALL, 0) != USB_SUCCESS) {
		dp->reg_data = NULL;
		goto err_unregister_client;
	}
#endif
#ifdef USBGEM_DEBUG_LEVEL
	usb_print_descr_tree(dp->dip, dp->reg_data);
#endif

	if (usbgem_open_pipes(dp) != USB_SUCCESS) {
		/* failed to open pipes */
		cmn_err(CE_WARN,
			"%s: usbgem_do_attach: failed to open pipes",
			dp->name);
		goto err_unregister_client;
	}

	/* 
	 * Initialize configuration
	 */
	dp->ugc = *gc;
	dp->tx_list = (struct txbuf *)((caddr_t)(&dp[1])+USBGEM_MCALLOC);

	dp->mtu		= ETHERMTU;
	dp->rxmode	= 0;
	dp->mii_fixedmode = B_FALSE;	/* default is autoneg */
	dp->speed	  = USBGEM_SPD_10;	/* default is 10Mbps */
	dp->full_duplex   = B_FALSE;	/* default is half */
	dp->flow_control  = FLOW_CONTROL_NONE;

	dp->nic_state = NIC_INACTIVE;
	dp->mii_state = MII_STATE_UNKNOWN;

        /* performance tuning parameters */
	dp->txthr    = ETHERMAX;	/* tx fifo threshoold */
	dp->txmaxdma = 16*4;		/* tx max dma burst size */
	dp->rxthr    = 128;		/* rx fifo threshoold */
	dp->rxmaxdma = 16*4;		/* rx max dma burst size */

	/* 
	 * allocate resources
	 */
	if (usbgem_alloc_memory(dp) != USB_SUCCESS) {
		goto err_close_pipes;
	}

	/* 
	 * Initialize mutexs and cv variables
	 */
	mutex_init(&dp->rxlock, NULL, MUTEX_DRIVER, NULL);
	mutex_init(&dp->txlock, NULL, MUTEX_DRIVER, NULL);
	cv_init(&dp->drain_cv, NULL, CV_DRIVER, NULL);

	/*
	 * Get media mode infomation from .conf file
	 */
	usbgem_read_conf(dp);

	/* XXX - buflen was aligned to support rtl8169 */
	dp->rx_buf_len = MAXPKTBUF(dp) + dp->ugc.usbgc_rx_header_len;

	/*
	 * Reset the chip
	 */
	if (USBGEM_RESET_CHIP(dp) != USB_SUCCESS) {
		cmn_err(CE_WARN,
			"%s: usbgem_do_attach: failed to reset the usb device",
			dp->name);
		goto err_destroy_locks;
	}

	/*
	 * HW dependant paremeter initialization
	 */
	if (USBGEM_ATTACH_CHIP(dp) != USB_SUCCESS) {
		cmn_err(CE_WARN,
			"%s: usbgem_do_attach: failed to attach the usb device",
			dp->name);
		goto err_destroy_locks;
	}

	/*
	 * Configure macinfo
	 */
	macinfo->gldm_devinfo	   = dip;
	macinfo->gldm_private	   = (caddr_t)dp;

	macinfo->gldm_reset        = usbgem_reset;
	macinfo->gldm_start        = usbgem_start;
	macinfo->gldm_stop         = usbgem_stop;
	macinfo->gldm_set_mac_addr = usbgem_set_mac_address;
	macinfo->gldm_send         = usbgem_send;
	macinfo->gldm_set_promiscuous = usbgem_set_promiscuous;
	macinfo->gldm_get_stats    = usbgem_get_stats;
	macinfo->gldm_ioctl        = NULL; 
	macinfo->gldm_set_multicast= usbgem_set_multicast;
	macinfo->gldm_intr         = NULL;
	macinfo->gldm_mctl         = NULL;

	macinfo->gldm_ident   = (char *)ddi_driver_name(dip);
	macinfo->gldm_type    = DL_ETHER;
	macinfo->gldm_minpkt  = 0;
	macinfo->gldm_maxpkt  = dp->mtu;
	macinfo->gldm_addrlen = ETHERADDRL;
	macinfo->gldm_saplen  = -2;
	macinfo->gldm_ppa     = dp->ugc.usbgc_ppa;

	cmn_err(CE_CONT,
	"!%s: %02x:%02x:%02x:%02x:%02x:%02x",
		dp->name,
		dp->dev_addr.ether_addr_octet[0],
		dp->dev_addr.ether_addr_octet[1],
		dp->dev_addr.ether_addr_octet[2],
		dp->dev_addr.ether_addr_octet[3],
		dp->dev_addr.ether_addr_octet[4],
		dp->dev_addr.ether_addr_octet[5]);

	/* copy mac address */
	dp->cur_addr = dp->dev_addr;
	macinfo->gldm_vendor_addr =
			dp->dev_addr.ether_addr_octet;
	macinfo->gldm_broadcast_addr =
			etherbroadcastaddr.ether_addr_octet;

	/* Init MII (scan phy)*/
	if (USBGEM_MII_INIT(dp) != USB_SUCCESS) {
		cmn_err(CE_WARN, "%s: usbgem_do_attach: mii_init failed",
			dp->name);
		goto err_destroy_locks;
	}

	/* reset mii and start mii link watcher */
	usbgem_mii_start(dp);

	/*
	 * Add interrupt to system.
	 */
	if (gld_register(dip, (char *)ddi_driver_name(dip), macinfo)
							!= DDI_SUCCESS) {
		cmn_err(CE_WARN,
			"%s: usbgem_do_attach: gld_register failed",
			dp->name);
		goto err_stop_mii;
	}

	if (usb_register_hotplug_cbs(dip, usbgem_disconnect_cb,
					usbgem_reconnect_cb) != USB_SUCCESS) {
		cmn_err(CE_WARN,
			"%s: usbgem_do_attach: failed to register hotplug cbs",
			dp->name);
		goto err_unregister;
	}

	DPRINTF(2, (CE_CONT, "!%s: usbgem_do_attach: return: success",
		dp->name));
	return dp;

err_unregister:
	gld_unregister(macinfo);

err_stop_mii:
	usbgem_mii_stop(dp);

err_destroy_locks:
	cv_destroy(&dp->drain_cv);
	mutex_destroy(&dp->txlock);
	mutex_destroy(&dp->rxlock);

err_free_memory:
	usbgem_free_memory(dp);

err_close_pipes:
	(void)usbgem_close_pipes(dp);

err_unregister_client:
#ifdef USB10
	usb_unregister_client(dp->dip, dp->reg_data);
#else
	usb_client_detach(dp->dip, dp->reg_data);
#endif
err_free_private:
	kmem_free((caddr_t)dp, USBGEM_LOCAL_DATA_SIZE(gc));
	gld_mac_free(macinfo);

	return NULL;
}

int
usbgem_do_detach(dev_info_t *dip)
{
	gld_mac_info_t		*macinfo;
	struct usbgem_dev	*dp;

	macinfo = (gld_mac_info_t *)ddi_get_driver_private(dip);
	dp = (struct usbgem_dev *)macinfo->gldm_private;

	usb_unregister_hotplug_cbs(dip);

	(void)gld_unregister(macinfo);

	usbgem_mii_stop(dp);

	mutex_destroy(&dp->txlock);
	mutex_destroy(&dp->rxlock);
	cv_destroy(&dp->drain_cv);

	if (dp->nic_state != NIC_DISCONNECTED) {
		(void)usbgem_free_memory(dp);
	}

	(void)usbgem_close_pipes(dp);
#ifdef USB10
	usb_unregister_client(dp->dip, dp->reg_data);
#else
	usb_client_detach(dp->dip, dp->reg_data);
#endif
	dp->reg_data = NULL;

	kmem_free((caddr_t)(dp->private), dp->priv_size);
	kmem_free((caddr_t)dp, USBGEM_LOCAL_DATA_SIZE(&dp->ugc));
	gld_mac_free(macinfo);

	DPRINTF(2, (CE_CONT, "!%s: usbgem_do_detach: return: success",
			ddi_driver_name(dip)));

	return DDI_SUCCESS;
}
